#
# Copyright (c) 2003 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# functions for splitting entities in a drawing

import math
import bisect

import Generic.layer
import Generic.point
import Generic.segment
import Generic.circle
import Generic.arc
import Generic.polyline

import intersections

def _segment_bounds(segment):
    _p1, _p2 = segment.getEndpoints()
    _x, _y = _p1.getCoords()
    _xmin = _xmax = _x
    _ymin = _ymax = _y
    _x, _y = _p2.getCoords()
    if _x < _xmin:
        _xmin = _x
    if _y < _ymin:
        _ymin = _y
    if _x > _xmax:
        _xmax = _x
    if _y > _ymax:
        _ymax = _y
    return _xmin, _ymin, _xmax, _ymax

def _circle_bounds(circle):
    _x, _y = circle.getCenter().getCoords()
    _rad = circle.getRadius()
    _xmin = _x - _rad
    _ymin = _y - _rad
    _xmax = _x + _rad
    _ymax = _y + _rad
    return _xmin, _ymin, _xmax, _ymax

def _no_overlap(r1, r2):
    return ((r1[2] < r2[0]) or # r1 xmax < r2 xmin
            (r1[0] > r2[2]) or # r1 xmin > r2 xmax
            (r1[3] < r2[1]) or # r1 ymax < r2 ymin
            (r1[1] > r2[3])) # r1 ymin > r2 ymax:

def _circle_seg_can_intersect(circle, seg):
    _val = False
    _cx, _cy = circle.getCenter().getCoords()
    _proj = seg.getProjection(_cx, _cy)
    if _proj is not None:
        _px, _py = _proj
        _sep = math.hypot((_cx - _px), (_cy - _py)) + 1e-10
        if _sep < circle.getRadius():
            _val = True
    return _val

def _circle_poly_can_intersect(circle, polyline):
    _val = False
    _cp = circle.getCenter()
    _radius = circle.getRadius()
    _dists = []
    for _pt in polyline.getPoints():
        _dists.append(_pt - _cp)
    _dmin = min(_dists) - 1e-10
    _dmax = max(_dists) + 1e-10
    if _dmin < _radius < _dmax:
        _val = True
    return _val

def split_segment(seg, pt):
    """Split a segment into two segments at a point.

split_segment(seg, pt)

seg: The segment to split
pt: The point used to split the segment.

There is presently no check to test that the point lies
on the segment.
    """
    if not isinstance(seg, Generic.segment.Segment):
        raise TypeError, "Invalid Segment: " + str(seg)
    if not isinstance(pt, Generic.point.Point):
        raise TypeError, "Invalid Point: " + str(pt)
    _p1, _p2 = seg.getEndpoints()
    _style = seg.getStyle()
    _linetype = seg.getLinetype()
    _color = seg.getColor()
    _thickness = seg.getThickness()
    _s1 = Generic.segment.Segment(_p1, pt, _style, _linetype, _color, _thickness)
    _s2 = Generic.segment.Segment(pt, _p2, _style, _linetype, _color, _thickness)
    return _s1, _s2

def _split_seg_2pts(seg, pt1, pt2):
    _p1, _p2 = seg.getEndpoints()
    _style = seg.getStyle()
    _linetype = seg.getLinetype()
    _color = seg.getColor()
    _thickness = seg.getThickness()
    _s1 = Generic.segment.Segment(_p1, pt1, _style, _linetype, _color, _thickness)
    _s2 = Generic.segment.Segment(pt1, pt2, _style, _linetype, _color, _thickness)
    _s3 = Generic.segment.Segment(pt2, _p2, _style, _linetype, _color, _thickness)
    return _s1, _s2, _s3

def split_circle(circle, pt):
    """Split a circle into a single arc.

split_circle(circle, pt)

circle: The circle to split
pt: The point used to determine the start/end angles of the arc.
    """
    if not isinstance(circle, Generic.circle.Circle):
        raise TypeError, "Invalid Circle: " + str(circle)
    if isinstance(circle, Generic.arc.Arc):
        raise TypeError, "Arc passed to split_circle()"
    if not isinstance(pt, Generic.point.Point):
        raise TypeError, "Invalid Point: " + str(pt)
    _cp = circle.getCenter()
    _rad = circle.getRadius()
    _cx, _cy = _cp.getCoords()
    _px, _py = pt.getCoords()
    _angle = (180.0/math.pi) * math.atan2((_py - _cy), (_px - _cx))
    if _angle < 0.0:
        _angle = _angle + 360.0
    return Generic.arc.Arc(_cp, _rad, _angle, _angle,
                           circle.getStyle(), circle.getLinetype(),
                           circle.getColor(), circle.getThickness())

def _split_circ_2pts(circ, p1, p2):
    _cp = circ.getCenter()
    _r = circ.getRadius()
    _cx, _cy = _cp.getCoords()            
    _px, _py = p1.getCoords()
    _angle1 = (180.0/math.pi) * math.atan2((_py - _cy),(_px - _cx))
    if _angle1 < 0.0:
        _angle1 = _angle1 + 360.0
    _px, _py = p2.getCoords()
    _angle2 = (180.0/math.pi) * math.atan2((_py - _cy),(_px - _cx))
    if _angle2 < 0.0:
        _angle2 = _angle2 + 360.0
    _style = circ.getStyle()
    _color = circ.getColor()
    _linetype = circ.getLinetype()
    _thickness = circ.getThickness()
    _a1 = Generic.arc.Arc(_cp, _r, _angle1, _angle2, _style,
                          _linetype, _color, _thickness)
    _a2 = Generic.arc.Arc(_cp, _r, _angle2, _angle1, _style,
                          _linetype, _color, _thickness)
    return _a1, _a2

def split_arc(arc, pt):
    """Split an arc into two connected arcs.

split_arc(arc, pt)

arc: The arc to split.
pt: The point used to determine the angle at which to split the arc.
    """
    if not isinstance(arc, Generic.arc.Arc):
        raise TypeError, "Invalid Arc: " + str(arc)
    if not isinstance(pt, Generic.point.Point):
        raise TypeError, "Invalid Point: " + str(pt)
    _cp = arc.getCenter()
    _cx, _cy = _cp.getCoords()
    _px, _py = pt.getCoords()
    _angle = (180.0/math.pi) * math.atan2((_py - _cy), (_px - _cx))
    if _angle < 0.0:
        _angle = _angle + 360.0
    if not arc.throughAngle(_angle):
        raise ValueError, "Arc does not exist at angle %g" % _angle
    _rad = arc.getRadius()
    _sa = arc.getStartAngle()
    _ea = arc.getEndAngle()
    _style = arc.getStyle()
    _linetype = arc.getLinetype()
    _color = arc.getColor()
    _thickness = arc.getThickness()
    _arc1 = Generic.arc.Arc(_cp, _rad, _sa, _angle,
                            _style, _linetype, _color, _thickness)
    _arc2 = Generic.arc.Arc(_cp, _rad, _angle, _ea,
                            _style, _linetype, _color, _thickness)
    return _arc1, _arc2

def _split_arc_2pts(arc, pt1, pt2):
    _cp = arc.getCenter()
    _rad = arc.getRadius()
    _sa = arc.getStartAngle()
    _ea = arc.getEndAngle()
    _cx, _cy = _cp.getCoords()
    _px, _py = pt1.getCoords()
    _angle1 = (180.0/math.pi) * math.atan2((_py - _cy), (_px - _cx))
    if _angle1 < 0.0:
        _angle1 = _angle1 + 360.0
    _px, _py = pt2.getCoords()
    _angle2 = (180.0/math.pi) * math.atan2((_py - _cy), (_px - _cx))
    if _angle2 < 0.0:
        _angle2 = _angle2 + 360.0
    if _sa < _ea:
        _a1 = min(_angle1, _angle2)
        _a2 = max(_angle1, _angle2)
    else:
        _d1 = _angle1 - _sa
        if _d1 < 0.0:
            _d1 = _angle1 + _sa
        _d2 = _angle2 - _sa
        if _d2 < 0.0:
            _d2 = _angle2 + _sa
        if _d1 < _d2:
            _a1 = _angle1
            _a2 = _angle2
        else:
            _a1 = _angle2
            _a2 = _angle1
    _style = arc.getStyle()
    _linetype = arc.getLinetype()
    _color = arc.getColor()
    _thickness = arc.getThickness()
    _arc1 = Generic.arc.Arc(_cp, _rad, _sa, _a1,
                            _style, _linetype, _color, _thickness)
    _arc2 = Generic.arc.Arc(_cp, _rad, _a1, _a2,
                            _style, _linetype, _color, _thickness)
    _arc3 = Generic.arc.Arc(_cp, _rad, _a2, _ea,
                            _style, _linetype, _color, _thickness)    
    return _arc1, _arc2, _arc3

def split_polyline(polyline, pt):
    if not isinstance(polyline, Generic.polyline.Polyline):
        raise TypeError, "Invalid Polyline: " + str(polyline)
    if not isinstance(pt, Generic.point.Point):
        raise TypeError, "Invalid Point: " + str(pt)
    _px, _py = pt.getCoords()
    _count = len(polyline)
    for _i in range(_count - 1):
        _hit = False
        _p1x, _p1y = polyline.getPoint(_i).getCoords()
        _p2x, _p2y = polyline.getPoint(_i + 1).getCoords()
        if abs(_p2x - _p1x) < 1e-10: # vertical
            if (abs(_p2x - _px) < 1e-10 and
                min(_p1y, _p2y) < _py < max(_p1y, _p2y)):
                _hit = True
        elif abs(_p2y - _p1y) < 1e-10: # horizontal
            if (abs(_p2y - _py) < 1e-10 and
                min(_p1x, _p2x) < _px < max(_p1x, _p2x)):
                _hit = True
        else:
            _slope = (_p2y - _p1y)/(_p2x - _p1x)
            _yint = _p2y - (_slope * _p2x)
            _ytest = (_slope * _px) + _yint
            if abs(_ytest - _py) < 1e-10:
                _hit = True
        if _hit:
            polyline.addPoint((_i + 1), pt)
            break
        
def _poly_poly_split(layer, objdict):
    _polylist = objdict['polylines']
    _objrects = objdict['objrects']
    while len(_polylist):
        _testpoly = _polylist.pop(0)
        if _testpoly not in _objrects:
            _objrects[_testpoly] = _testpoly.getBounds()
        _hit_test = False            
        for _poly in _polylist:
            if _poly not in _objrects:
                _objrects[_poly] = _poly.getBounds()
            if _no_overlap(_objrects[_testpoly], _objrects[_poly]):
                continue
            _maxi = len(_testpoly) - 1
            for _i in range(_maxi):
                _p1 = _testpoly.getPoint(_i)
                _p2 = _testpoly.getPoint(_i + 1)
                _x1, _y1 = _p1.getCoords()
                _x2, _y2 = _p2.getCoords()
                _maxj = len(_poly) - 1
                for _j in range(_maxj):
                    _p3 = _poly.getPoint(_j)
                    _p4 = _poly.getPoint(_j + 1)
                    if _p1 == _p3 or _p1 == _p4 or _p2 == _p3 or _p2 == _p4:
                        continue
                    _d = intersections.denom(_p1, _p2, _p3, _p4)
                    if abs(_d) > 1e-10: # NOT parallel
                        _rn = intersections.rnum(_p1, _p2, _p3, _p4)
                        if abs(_rn) > 1e-10: # NOT colinear
                            _sn = intersections.snum(_p1, _p2, _p3 ,_p4)
                            _r = _rn/_d
                            _s = _sn/_d
                            if ((0.0 < _r < 1.0) and (0.0 < _s < 1.0)):
                                _xi = _x1 + _r * (_x2 - _x1)
                                _yi = _y1 + _r * (_y2 - _y1)
                                _lp = layer.find('point', _xi, _yi)
                                if _lp is None:
                                    _lp = Generic.point.Point(_xi, _yi)
                                    layer.addObject(_lp)
                                _hit_test = True
                                _testpoly.addPoint((_i + 1), _lp)
                                _poly.addPoint((_j + 1), _lp)
                                break
                if _hit_test:
                    break
            if _hit_test:
                break
        if _hit_test:
            _polylist.insert(0, _testpoly)
    
def _arc_polyline_split(layer, objdict):
    _arclist = objdict['arcs']
    _arclist.sort()
    _polylist = objdict['polylines']
    _objstate = {}
    _newarcs = []
    _oldarcs = []
    _objrects = objdict['objrects']
    for _arc in _arclist:
        _objstate[id(_arc)] = True
        _oldarcs.append(_arc)
    while len(_arclist):
        _testarc = _arclist.pop(0)
        _testid = id(_testarc)
        if _testid in _objstate and not _objstate[_testid]:
            continue
        if _testid not in _objstate:
            _objstate[_testid] = True
        if _testid not in _objrects:
            _objrects[_testid] = _testarc.getBounds()
        _ep1, _ep2 = _testarc.getEndpoints()            
        for _poly in _polylist:
            _polyid = id(_poly)
            if _polyid not in _objrects:
                _objrects[_polyid] = _poly.getBounds()
            if _no_overlap(_objrects[_testid], _objrects[_polyid]):
                continue
            _hit = False
            _ptcount = len(_poly) - 1
            for _i in range(_ptcount):
                _lp1 = _poly.getPoint(_i)
                _lp2 = _poly.getPoint(_i + 1)
                _tempseg = Generic.segment.Segment(_lp1, _lp2)
                _ipts = intersections.find_intersections(_testarc, _tempseg)
                if len(_ipts):
                    _arc1 = _arc2 = _arc3 = None
                    _count = len(_ipts)
                    if _count == 1:
                        _x, _y = _ipts[0]
                        _ip = layer.find('point', _x, _y)
                        if _ip is None:
                            _ip = Generic.point.Point(_x, _y)
                            layer.addObject(_ip)
                        if _ip != _ep1 and _ip != _ep2:
                            _hit = True
                            _arc1, _arc2 = split_arc(_testarc, _ip)
                        if _ip != _lp1 and _ip != _lp2:
                            _hit = True
                            _poly.addPoint((_i + 1), _ip)
                    elif _count == 2:
                        _x, _y = _ipts[0]
                        _ip1 = layer.find('point', _x, _y)
                        if _ip1 is None:
                            _ip1 = Generic.point.Point(_x, _y)
                            layer.addObject(_ip1)
                        _x, _y = _ipts[1]
                        _ip2 = layer.find('point', _x, _y)
                        if _ip2 is None:
                            _ip2 = Generic.point.Point(_x, _y)
                            layer.addObject(_ip2)
                        if (_ip1 != _ep1 and
                            _ip1 != _ep2 and
                            _ip2 != _ep1 and
                            _ip2 != _ep2):
                            _hit = True
                            _arc1, _arc2, _arc3 = _split_arc_2pts(_testarc, _ip1, _ip2)
                        elif _ip1 == _ep1 or _ip1 == _ep2:
                            if _ip2 != _ep1 and _ip2 != _ep2:
                                _hit = True
                                _arc1, _arc2 = split_arc(_testarc, _ip2)
                        elif _ip2 == _ep1 or _ip2 == _ep2:
                            if _ip1 != _ep1 and _ip1 != _ep2:
                                _hit = True
                                _arc1, _arc2 = split_arc(_testarc, _ip1)
                        _d1 = _ip1 - _lp1
                        _d2 = _lp2 - _lp1
                        if _d1 < _d2:
                            _p1 = _ip1
                            _p2 = _ip2
                        else:
                            _p1 = _ip2
                            _p2 = _ip1
                        if (_p1 != _lp1 and _p1 != _lp2):
                            _hit = True
                            _poly.addPoint((_i + 1), _p1)
                            _i = _i + 1
                        if (_p2 != _lp1 and _p2 != _lp2):
                            _hit = True
                            _poly.addPoint((_i + 1), _p2)
                    else:
                        raise ValueError, "Unexpected arc/polyline count: %d" % _count
                    #
                    # if _arc1 is not None then _testarc was split
                    #
                    if _arc1 is not None:
                        _objstate[_testid] = False
                        _newarcs.append(_arc1)
                        _objstate[id(_arc1)] = True
                        bisect.insort_right(_arclist, _arc1)
                    if _arc2 is not None:
                        _newarcs.append(_arc2)
                        _objstate[id(_arc2)] = True
                        bisect.insort_right(_arclist, _arc2)
                    if _arc3 is not None:
                        _newarcs.append(_arc3)
                        _objstate[id(_arc3)] = True
                        bisect.insort_right(_arclist, _arc3)
                    if _hit:
                        break
            if _hit:
                break
    _layerarclist = []
    for _arc in _newarcs:
        if _objstate[id(_arc)]:
            layer.addObject(_arc)
            _layerarclist.append(_arc)
        else:
            _cp = _arc.getCenter()
            _cp.freeUser(_arc)
    for _arc in _oldarcs:
        if _objstate[id(_arc)]:
            _layerarclist.append(_arc)
    objdict['arcs'] = _layerarclist
    for _arc in _oldarcs:
        if not _objstate[id(_arc)]:
            layer.delObject(_arc)

def _arc_arc_split(layer, objdict):
    _arclist = objdict['arcs']
    _arclist.sort()
    _arcstate = {}
    _newarcs = []
    _oldarcs = []
    _objrects = objdict['objrects']
    for _arc in _arclist:
        _arcstate[id(_arc)] = True
        _oldarcs.append(_arc)
    while len(_arclist):
        _testarc = _arclist.pop(0)
        _testid = id(_testarc)
        if _testid in _arcstate and not _arcstate[_testid]:
            continue
        if _testid not in _arcstate:
            _arcstate[_testid] = True
        if _testid not in _objrects:
            _objrects[_testid] = _testarc.getBounds()
        _ep1, _ep2 = _testarc.getEndpoints()
        for _arc in _arclist:
            _arcid = id(_arc)
            if _arcid in _arcstate and not _arcstate[_arcid]:
                continue
            if _arcid not in _arcstate:
                _arcstate[_arcid] = True
            if _arcid not in _objrects:
                _objrects[_arcid] = _arc.getBounds()
            if _no_overlap(_objrects[_testid], _objrects[_arcid]):
                continue
            _ep3, _ep4 = _arc.getEndpoints()
            _ipts = intersections.find_intersections(_testarc, _arc)
            if len(_ipts):
                _a1 = _a2 = _a3 = _a4 = _a5 = _a6 = None
                _count = len(_ipts)
                if _count == 1:
                    _x, _y = _ipts[0]
                    _ip = layer.find('point', _x, _y)
                    if _ip is None:
                        _ip = Generic.point.Point(_x, _y)
                        layer.addObject(_ip)
                    if _ip != _ep1 and _ip != _ep2:
                        _a1, _a2 = split_arc(_testarc, _ip)
                    if _ip != _ep3 and _ip != _ep4:
                        _a4, _a5 = split_arc(_arc, _ip)
                elif _count == 2:
                    _x, _y = _ipts[0]
                    _ip1 = layer.find('point', _x, _y)
                    if _ip1 is None:
                        _ip1 = Generic.point.Point(_x, _y)
                        layer.addObject(_ip1)
                    _x, _y = _ipts[1]
                    _ip2 = layer.find('point', _x, _y)
                    if _ip2 is None:
                        _ip2 = Generic.point.Point(_x, _y)
                        layer.addObject(_ip2)
                    if (_ip1 != _ep1 and
                        _ip1 != _ep2 and
                        _ip2 != _ep1 and
                        _ip2 != _ep2):
                        _a1, _a2, _a3 = _split_arc_2pts(_testarc, _ip1, _ip2)
                    elif _ip1 == _ep1 or _ip1 == _ep2:
                        if _ip2 != _ep1 and _ip2 != _ep2:
                            _a1, _a2 = split_arc(_testarc, _ip2)
                    elif _ip2 == _ep1 or _ip2 == _ep2:
                        if _ip1 != _ep1 and _ip1 != _ep2:
                            _a1, _a2 = split_arc(_testarc, _ip1)
                    if (_ip1 != _ep3 and
                        _ip1 != _ep4 and
                        _ip2 != _ep3 and
                        _ip2 != _ep4):
                        _a4, _a5, _a6 = _split_arc_2pts(_arc, _ip1, _ip2)
                    elif _ip1 == _ep3 or _ip1 == _ep4:
                        if _ip2 != _ep3 and _ip2 != _ep4:
                            _a4, _a5 = split_arc(_arc, _ip2)
                    elif _ip2 == _ep3 or _ip2 == _ep4:
                        if _ip1 != _ep3 and _ip1 != _ep4:
                            _a4, _a5 = split_arc(_arc, _ip1)
                else:
                    raise ValueError, "Unexpected arc/arc int count: %d" % _count
                #
                # if _a1 is not None then _testarc was split
                #
                if _a1 is not None:
                    _arcstate[_testid] = False
                    bisect.insort_right(_arclist, _a1)
                    _newarcs.append(_a1)
                    _arcstate[id(_a1)] = True
                if _a2 is not None:
                    bisect.insort_right(_arclist, _a2)
                    _newarcs.append(_a2)
                    _arcstate[id(_a2)] = True
                if _a3 is not None:
                    bisect.insort_right(_arclist, _a3)
                    _newarcs.append(_a3)
                    _arcstate[id(_a3)] = True
                #
                # if _a4 is not None then _arc was split
                #
                if _a4 is not None:
                    _arcstate[_arcid] = False
                    bisect.insort_right(_arclist, _a4)
                    _newarcs.append(_a4)
                    _arcstate[id(_a4)] = True
                if _a5 is not None:
                    bisect.insort_right(_arclist, _a5)
                    _newarcs.append(_a5)
                    _arcstate[id(_a5)] = True
                if _a6 is not None:
                    bisect.insort_right(_arclist, _a6)
                    _newarcs.append(_a6)
                    _arcstate[id(_a6)] = True
                break
    _layerarclist = []
    for _arc in _newarcs:
        if _arcstate[id(_arc)]:
            layer.addObject(_arc)
            _layerarclist.append(_arc)
        else:
            _cp = _arc.getCenter()
            _cp.freeUser(_arc)
    for _arc in _oldarcs:
        if _arcstate[id(_arc)]:
            _layerarclist.append(_arc)
    objdict['arcs'] = _layerarclist
    for _arc in _oldarcs:
        if not _arcstate[id(_arc)]:
            layer.delObject(_arc)

def _circ_polyline_split(layer, objdict):
    _circlist = objdict['circles']
    _circlist.sort()
    _polylist = objdict['polylines']
    _objstate = {}
    _newarcs = []
    _oldcircs = []
    _objrects = objdict['objrects']
    for _circ in _circlist:
        _objstate[id(_circ)] = True
        _oldcircs.append(_circ)
    while len(_circlist):
        _testcirc = _circlist.pop(0)
        _testid = id(_testcirc)
        if _testid in _objstate and not _objstate[_testid]:
            continue
        if _testid not in _objstate:
            _objstate[_testid] = True
        if _testid not in _objrects:
            _objrects[_testid] = _circle_bounds(_testcirc)
        for _poly in _polylist:
            _polyid = id(_poly)
            if _polyid not in _objrects:
                _objrects[_polyid] = _poly.getBounds()
            if _no_overlap(_objrects[_testid], _objrects[_polyid]):
                continue
            _count = len(_poly) - 1
            _hit = False
            for _i in range(_count):
                _lp1 = _poly.getPoint(_i)
                _lp2 = _poly.getPoint(_i + 1)
                _tempseg = Generic.segment.Segment(_lp1, _lp2)
                _ipts = intersections.find_intersections(_testcirc, _tempseg)
                if len(_ipts):
                    _hit = True
                    _arc1 = _arc2 = None
                    _count = len(_ipts)
                    if _count == 1:
                        _x, _y = _ipts[0]
                        _ip = layer.find('point', _x, _y)
                        if _ip is None:
                            _ip = Generic.point.Point(_x, _y)
                            layer.addObject(_ip)
                        _arc1 = split_circle(_testcirc, _ip)
                        if _ip != _lp1 and _ip != _lp2:
                            _poly.addPoint((_i + 1), _ip)
                    elif _count == 2:
                        _x, _y = _ipts[0]
                        _ip1 = layer.find('point', _x, _y)
                        if _ip1 is None:
                            _ip1 = Generic.point.Point(_x, _y)
                            layer.addObject(_ip1)
                        _x, _y = _ipts[1]
                        _ip2 = layer.find('point', _x, _y)
                        if _ip2 is None:
                            _ip2 = Generic.point.Point(_x, _y)
                            layer.addObject(_ip2)
                        _arc1, _arc2 = _split_circ_2pts(_circ, _ip1, _ip2)
                        _d1 = _ip1 - _lp1
                        _d2 = _ip2 - _lp1
                        if _d1 < _d2:
                            _p1 = _ip1
                            _p2 = _ip2
                        else:
                            _p1 = _ip2
                            _p2 = _ip1
                        if (_p1 != _lp1 and _p1 != _lp2):
                            _poly.addPoint((_i + 1), _p1)
                            _i = _i + 1
                        if (_p2 != _lp1 and _p2 != _lp2):
                            _poly.addPoint((_i + 1), _p2)
                    else:
                        raise ValueError, "Unexpected circle/polyline count: %d" % _count
                    #
                    # if _arc1 is not None the circle was split
                    #
                    if _arc1 is not None:
                        _objstate[_testid] = False
                        _newarcs.append(_arc1)
                        _objstate[id(_arc1)] = True
                    if _arc2 is not None:
                        _newarcs.append(_arc2)
                        _objstate[id(_arc2)] = True
                    break
            if _hit:
                break
    if len(_newarcs):
        _arclist = objdict['arcs']
        for _arc in _newarcs:
            layer.addObject(_arc)
            _arclist.append(_arc)
    _layercirclelist = []
    for _circ in _oldcircs:
        if _objstate[id(_circ)]:
            _layercirclelist.append(_circ)
    objdict['circles'] = _layercirclelist
    for _circ in _oldcircs:
        if not _objstate[id(_circ)]:
            layer.delObject(_circ)

def _circ_arc_split(layer, objdict):
    _circlist = objdict['circles']
    _circlist.sort()
    _oldcircs = []
    _arclist = objdict['arcs']
    _arclist.sort()
    _objstate = {}
    _oldarcs = []
    _newarcs = []
    _objrects = objdict['objrects']
    for _circ in _circlist:
        _objstate[id(_circ)] = True
        _oldcircs.append(_circ)
    for _arc in _arclist:
        _objstate[id(_arc)] = True
        _oldarcs.append(_arc)
    while len(_circlist):
        _testcirc = _circlist.pop(0)
        _testid = id(_testcirc)
        if _testid in _objstate and not _objstate[_testid]:
            continue
        if _testid not in _objstate:
            _objstate[_testid] = True
        if _testid not in _objrects:
            _objrects[_testid] = _circle_bounds(_testcirc)
        for _arc in _arclist:
            _arcid = id(_arc)
            if _arcid in _objstate and not _objstate[_arcid]:
                continue
            if _arcid not in _objstate:
                _objstate[_arcid] = True
            if _arcid not in _objrects:
                _objrects[_arcid] = _arc.getBounds()
            if _no_overlap(_objrects[_arcid], _objrects[_testid]):
                continue
            _ipts = intersections.find_intersections(_testcirc, _arc)
            if len(_ipts):
                _ep1, _ep2 = _arc.getEndpoints()
                _ca1 = _ca2 = _arc1 = _arc2 = _arc3 = None
                _count = len(_ipts)
                if _count == 1:
                    _x, _y = _ipts[0]
                    _ip = layer.find('point', _x, _y)
                    if _ip is None:
                        _ip = Generic.point.Point(_x, _y)
                        layer.addObject(_ip)
                    _ca1 = split_circle(_testcirc, _ip)
                    if _ip != _ep1 and _ip != _ep2:
                        _arc1, _arc2 = split_arc(_arc, _ip)
                elif _count == 2:
                    _x, _y = _ipts[0]
                    _ip1 = layer.find('point', _x, _y)
                    if _ip1 is None:
                        _ip1 = Generic.point.Point(_x, _y)
                        layer.addObject(_ip1)
                    _x, _y = _ipts[1]
                    _ip2 = layer.find('point', _x, _y)
                    if _ip2 is None:
                        _ip2 = Generic.point.Point(_x, _y)
                        layer.addObject(_ip2)
                    _ca1, _ca2 = _split_circ_2pts(_testcirc, _ip1, _ip2)
                    if (_ip1 != _ep1 and
                        _ip1 != _ep2 and
                        _ip2 != _ep1 and
                        _ip2 != _ep2):
                        _arc1, _arc2, _arc3 = _split_arc_2pts(_arc, _ip1, _ip2)
                    elif _ip1 == _ep1 or _ip1 == _ep2:
                        if _ip2 != _ep1 and _ip2 != _ep2:
                            _arc1, _arc2 = split_arc(_arc, _ip2)
                    elif _ip2 == _ep1 or _ip2 == _ep2:
                        if _ip1 !=_ep1 and _ip1 != _ep2:
                            _arc1, _arc2 = split_arc(_arc, _ip1)
                else:
                    raise ValueError, "Unexpected circle/arc count: %d" % _count
                #
                # if _ca1 is not none then the circle was split
                #
                if _ca1 is not None:
                    _objstate[_testid] = False
                    bisect.insort_right(_arclist, _ca1)
                    _newarcs.append(_ca1)
                    _objstate[id(_ca1)] = True
                if _ca2 is not None:
                    bisect.insort_right(_arclist, _ca2)
                    _newarcs.append(_ca2)
                    _objstate[id(_ca2)] = True
                #
                # if _arc1 is not None then the arc was split
                #
                if _arc1 is not None:
                    _objstate[_arcid] = False
                    bisect.insort_right(_arclist, _arc1)
                    _newarcs.append(_arc1)
                    _objstate[id(_arc1)] = True
                if _arc2 is not None:
                    bisect.insort_right(_arclist, _arc2)
                    _newarcs.append(_arc2)
                    _objstate[id(_arc2)] = True
                if _arc3 is not None:
                    bisect.insort_right(_arclist, _arc3)
                    _newarcs.append(_arc3)
                    _objstate[id(_arc3)] = True
                break
    _layerarclist = []
    for _arc in _newarcs:
        if _objstate[id(_arc)]:
            layer.addObject(_arc)
            _layerarclist.append(_arc)
        else:
            _cp = _arc.getCenter()
            _cp.freeUser(_arc)
    for _arc in _oldarcs:
        if _objstate[id(_arc)]:
            _layerarclist.append(_arc)
    objdict['arcs'] = _layerarclist
    _layercirclelist = []
    for _circ in _oldcircs:
        if _objstate[id(_circ)]:
            _layercirclelist.append(_circ)
    objdict['circles'] = _layercirclelist
    for _arc in _oldarcs:
        if not _objstate[id(_arc)]:
            layer.delObject(_arc)
    for _circ in _oldcircs:
        if not _objstate[id(_circ)]:
            layer.delObject(_circ)

def _circ_circ_split(layer, objdict):
    print "in _circ_circ_split() ..."
    _circlist = objdict['circles']
    _circlist.sort()
    _circstate = {}
    _oldcircs = []
    _newarcs = []
    _objrects = objdict['objrects']
    for _circ in _circlist:
        _circstate[id(_circ)] = True
        _oldcircs.append(_circ)
    while len(_circlist):
        _testcirc = _circlist.pop(0)
        _testid = id(_testcirc)
        if _testid in _circstate and not _circstate[_testid]:
            continue
        if _testid not in _circstate:
            _circstate[_testid] = True
        if _testid not in _objrects:
            _objrects[_testid] = _circle_bounds(_testcirc)
        for _circ in _circlist:
            _circid = id(_circ)
            if _circid in _circstate and not _circstate[_circid]:
                continue
            if _circid not in _circstate:
                _circstate[_circid] = True
            if _circid not in _objrects:
                _objrects[_circid] = _circle_bounds(_circ)
            if _no_overlap(_objrects[_testid], _objrects[_circid]):
                continue
            _ipts = intersections.find_intersections(_testcirc, _circ)
            if len(_ipts):
                _count = len(_ipts)
                if _count == 1:
                    _x, _y = _ipts[0]
                    _ip = layer.find('point', _x, _y)
                    if _ip is None:
                        _ip = Generic.point.Point(_x, _y)
                        layer.addObject(_ip)
                    _circstate[_testid] = False
                    _arc = split_circle(_testcirc, _ip)
                    _newarcs.append(_arc)
                    _circstate[_circid] = False
                    _arc = split_circle(_circ, _ip)
                    _newarcs.append(_arc)
                elif _count == 2:
                    _x, _y = _ipts[0]
                    _ip1 = layer.find('point', _x, _y)
                    if _ip1 is None:
                        _ip1 = Generic.point.Point(_x, _y)
                        layer.addObject(_ip1)
                    _x, _y = _ipts[1]
                    _ip2 = layer.find('point', _x, _y)
                    if _ip2 is None:
                        _ip2 = Generic.point.Point(_x, _y)
                        layer.addObject(_ip2)
                    _arc1, _arc2 = _split_circ_2pts(_testcirc, _ip1, _ip2)
                    _circstate[_testid] = False
                    _newarcs.append(_arc1)
                    _newarcs.append(_arc2)
                    _circstate[_circid] = False
                    _arc1, _arc2 = _split_circ_2pts(_circ, _ip1, _ip2)
                    _newarcs.append(_arc1)
                    _newarcs.append(_arc2)
                else:
                    raise ValueError, "Unexpected circle/circle count: %d" % _count
                break
    if len(_newarcs):
        _arclist = objdict['arcs']
        for _arc in _newarcs:
            layer.addObject(_arc)
            _arclist.append(_arc)
    _layercirclelist = []
    for _circ in _oldcircs:
        if _circstate[id(_circ)]:
            _layercirclelist.append(_circ)            
    objdict['circles'] = _layercirclelist
    for _circ in _oldcircs:
        if not _circstate[id(_circ)]:
            layer.delObject(_circ)

def _seg_polyline_split(layer, objdict):
    _seglist = objdict['segments']
    _seglist.sort()
    _polylist = objdict['polylines']
    _segstate = {}
    _newsegs = []
    _oldsegs = []
    _objrects = objdict['objrects']
    for _seg in _seglist:
        _segstate[id(_seg)] = True
        _oldsegs.append(_seg)
    while len(_seglist):
        _testseg = _seglist.pop(0)
        _testid = id(_testseg)
        if _testid in _segstate and not _segstate[_testid]:
            continue
        if _testid not in _segstate:
            _segstate[_testid] = True
        _p1, _p2 = _testseg.getEndpoints()
        if _testid not in _objrects:
            _objrects[_testid] = _segment_bounds(_testseg)
        for _poly in _polylist:
            _polyid = id(_poly)
            if _polyid not in _objrects:
                _objrects[_polyid] = _poly.getBounds()
            if _no_overlap(_objrects[_testid], _objrects[_polyid]):
                continue
            _count = len(_poly) - 1
            _hit = False
            for _i in range(_count):
                _lp1 = _poly.getPoint(_i)
                _lp2 = _poly.getPoint(_i + 1)
                if _p1 == _lp1 or _p1 == _lp2 or _p2 == _lp1 or _p2 == _lp2:
                    continue
                _tempseg = Generic.segment.Segment(_lp1, _lp2)
                _ipts = intersections.find_intersections(_testseg, _tempseg)
                if len(_ipts):
                    assert len(_ipts) == 1, "Invalid length: %d" % len(_ipts)
                    _hit = True
                    _x, _y = _ipts[0]
                    _ip = layer.find('point', _x, _y)
                    if _ip is None:
                        _ip = Generic.point.Point(_x, _y)
                        layer.addObject(_ip)
                    _s1, _s2 = split_segment(_testseg, _ip)
                    _segstate[_testid] = False
                    _newsegs.append(_s1)
                    _segstate[id(_s1)] = True
                    _newsegs.append(_s2)
                    _segstate[id(_s2)] = True
                    bisect.insort_right(_seglist, _s1)
                    bisect.insort_right(_seglist, _s2)
                    _poly.addPoint((_i + 1), _ip)
                    break
            if _hit:
                break
    _layerseglist = []
    for _seg in _newsegs:
        if _segstate[id(_seg)]:
            layer.addObject(_seg)
            _layerseglist.append(_seg)
        else:
            _p1, _p2 = _seg.getEndpoints()
            _p1.freeUser(_seg)
            _p2.freeUser(_seg)
    for _seg in _oldsegs:
        if _segstate[id(_seg)]:
            _layerseglist.append(_seg)
    objdict['segments'] = _layerseglist
    for _seg in _oldsegs:
        if not _segstate[id(_seg)]:
            layer.delObject(_seg)

def _seg_arc_split(layer, objdict):
    _seglist = objdict['segments']
    _seglist.sort()
    _arclist = objdict['arcs']
    _arclist.sort()
    _objstate = {}
    _newsegs = []
    _oldsegs = []
    _newarcs = []
    _oldarcs = []
    _objrects = objdict['objrects']
    for _seg in _seglist:
        _objstate[id(_seg)] = True
        _oldsegs.append(_seg)
    for _arc in _arclist:
        _objstate[id(_arc)] = True
        _oldarcs.append(_arc)
    while len(_seglist):
        _testseg = _seglist.pop(0)
        _testid = id(_testseg)
        if _testid in _objstate and not _objstate[_testid]:
            continue
        if _testid not in _objstate:
            _objstate[_testid] = True
        _p1, _p2 = _testseg.getEndpoints()
        if _testid not in _objrects:
            _objrects[_testid] = _segment_bounds(_testseg)
        for _arc in _arclist:
            _arcid = id(_arc)
            if _arcid in _objstate and not _objstate[_arcid]:
                continue
            if _arcid not in _objstate:
                _objstate[_arcid] = True
            if _arcid not in _objrects:
                _objrects[_arcid] = _arc.getBounds()
            if _no_overlap(_objrects[_arcid], _objrects[_testid]):
                continue
            _ipts = intersections.find_intersections(_testseg, _arc)
            if len(_ipts):
                _s1 = _s2 = _s3 = _arc1 = _arc2 = _arc3 = None
                _p1, _p2 = _testseg.getEndpoints()
                _ep1, _ep2 = _arc.getEndpoints()
                _count = len(_ipts)
                if _count == 1:
                    _x, _y = _ipts[0]
                    _ip = layer.find('point', _x, _y)
                    if _ip is None:
                        _ip = Generic.point.Point(_x, _y)
                        layer.addObject(_ip)
                    if _ip != _ep1 and _ip != _ep2:
                        _arc1, _arc2 = split_arc(_arc, _ip)
                    if _ip != _p1 and _ip != _p2:
                        _s1, _s2 = split_segment(_testseg, _ip)
                elif _count == 2:
                    _x, _y = _ipts[0]
                    _ip1 = layer.find('point', _x, _y)
                    if _ip1 is None:
                        _ip1 = Generic.point.Point(_x, _y)
                        layer.addObject(_ip1)
                    _x, _y = _ipts[1]
                    _ip2 = layer.find('point', _x, _y)
                    if _ip2 is None:
                        _ip2 = Generic.point.Point(_x, _y)
                        layer.addObject(_ip2)
                    if (_ip1 != _ep1 and
                        _ip1 != _ep2 and
                        _ip2 != _ep1 and
                        _ip2 != _ep2):
                        _arc1, _arc2, _arc3 = _split_arc_2pts(_arc, _ip1, _ip2)
                    elif _ip1 == _ep1 or _ip1 == _ep2:
                        if _ip2 != _ep1 and _ip2 != _ep2:
                            _arc1, _arc2 = split_arc(_arc, _ip2)
                    elif _ip2 == _ep1 or _ip2 == _ep2:
                        if _ip1 !=_ep1 and _ip1 != _ep2:
                            _arc1, _arc2 = split_arc(_arc, _ip1)
                    if (_ip1 != _p1 and
                        _ip1 != _p2 and
                        _ip2 != _p1 and
                        _ip2 != _p2):
                        if ((_p1 - _ip1) < (_p1 - _ip2)):
                            _m1 = _ip1
                            _m2 = _ip2
                        else:
                            _m1 = _ip2
                            _m2 = _ip1
                        _s1, _s2, _s3 = _split_seg_2pts(_testseg, _m1, _m2)
                    elif _ip1 == _p1 or _ip1 == _p2:
                        if _ip2 != _p1 and _ip2 != _p2:
                            _s1, _s2 = split_segment(_testseg, _ip2)
                    elif _ip2 == _p1 or _ip2 == _p2:
                        if _ip1 != _p1 and _ip1 != _p2:
                            _s1, _s2 = split_segment(_testseg, _ip1)
                else:
                    raise ValueError, "Unexpected seg/arc count: %d" % _count
                #
                # if _s1 is not None then the segment was split
                #
                if _s1 is not None:
                    _objstate[_testid] = False
                    _newsegs.append(_s1)
                    _objstate[id(_s1)] = True
                    bisect.insort_right(_seglist, _s1)
                if _s2 is not None:
                    _newsegs.append(_s2)
                    _objstate[id(_s2)] = True
                    bisect.insort_right(_seglist, _s2) 
                if _s3 is not None:
                    _newsegs.append(_s3)
                    _objstate[id(_s3)] = True
                    bisect.insort_right(_seglist, _s3)
                #
                # if _arc1 is not None then the arc was split
                #
                if _arc1 is not None:
                    _objstate[_arcid] = False
                    _newarcs.append(_arc1)
                    _objstate[id(_arc1)] = True
                    bisect.insort_right(_arclist, _arc1)
                if _arc2 is not None:
                    _newarcs.append(_arc2)
                    _objstate[id(_arc2)] = True
                    bisect.insort_right(_arclist, _arc2)
                if _arc3 is not None:
                    _newarcs.append(_arc3)
                    _objstate[id(_arc3)] = True
                    bisect.insort_right(_arclist, _arc3)
                break
    _layerarclist = []
    for _arc in _newarcs:
        if _objstate[id(_arc)]:
            layer.addObject(_arc)
            _layerarclist.append(_arc)
        else:
            _cp = _arc.getCenter()
            _cp.freeUser(_arc)
    for _arc in _oldarcs:
        if _objstate[id(_arc)]:
            _layerarclist.append(_arc)
    objdict['arcs'] = _layerarclist
    _layerseglist = []
    for _seg in _newsegs:
        if _objstate[id(_seg)]:
            layer.addObject(_seg)
            _layerseglist.append(_seg)
        else:
            _p1, _p2 = _seg.getEndpoints()
            _p1.freeUser(_seg)
            _p2.freeUser(_seg)
    for _seg in _oldsegs:
        if _objstate[id(_seg)]:
            _layerseglist.append(_seg)
    objdict['segments'] = _layerseglist
    for _arc in _oldarcs:
        if not _objstate[id(_arc)]:
            layer.delObject(_arc)
    for _seg in _oldsegs:
        if not _objstate[id(_seg)]:
            layer.delObject(_seg)

def _seg_circle_split(layer, objdict):
    _seglist = objdict['segments']
    _seglist.sort()
    _circlist = objdict['circles']
    _circlist.sort()
    _objstate = {}
    _newsegs = []
    _oldsegs = []
    _oldcircs = []
    _newarcs = []    
    _objrects = objdict['objrects']
    for _seg in _seglist:
        _objstate[id(_seg)] = True
        _oldsegs.append(_seg)
    for _circ in _circlist:
        _objstate[id(_circ)] = True
        _oldcircs.append(_circ)
    while len(_seglist):
        _testseg = _seglist.pop(0)
        _testid = id(_testseg)
        if _testid in _objstate and not _objstate[_testid]:
            continue
        if _testid not in _objstate:
            _objstate[_testid] = True
        _p1, _p2 = _testseg.getEndpoints()
        if _testid not in _objrects:
            _objrects[_testid] = _segment_bounds(_testseg)
        for _circ in _circlist:
            _circid = id(_circ)
            if _circid in _objstate and not _objstate[_circid]:
                continue
            if _circid not in _objrects:
                _objrects[_circid] = _circle_bounds(_circ)
            if _no_overlap(_objrects[_circid], _objrects[_testid]):
                continue
            _ipts = intersections.find_intersections(_testseg, _circ)
            if len(_ipts):
                _s1 = _s2 = _s3 = _arc1 = _arc2 = None
                _p1, _p2 = _testseg.getEndpoints()
                _count = len(_ipts)
                if _count == 1:
                    _x, _y = _ipts[0]
                    _ip = layer.find('point', _x, _y)
                    if _ip is None:
                        _ip = Generic.point.Point(_x, _y)
                        layer.addObject(_ip)
                    _arc1 = split_circle(_circ, _ip)
                    if _ip != _p1 and _ip != _p2:
                        _s1, _s2 = split_segment(_testseg, _ip)
                elif _count == 2:
                    _x, _y = _ipts[0]
                    _ip1 = layer.find('point', _x, _y)
                    if _ip1 is None:
                        _ip1 = Generic.point.Point(_x, _y)
                        layer.addObject(_ip1)
                    _x, _y = _ipts[1]
                    _ip2 = layer.find('point', _x, _y)
                    if _ip2 is None:
                        _ip2 = Generic.point.Point(_x, _y)
                        layer.addObject(_ip2)
                    _arc1, _arc2 = _split_circ_2pts(_circ, _ip1, _ip2)
                    if (_ip1 != _p1 and
                        _ip1 != _p2 and
                        _ip2 != _p1 and
                        _ip2 != _p2):
                        if ((_p1 - _ip1) < (_p1 - _ip2)):
                            _m1 = _ip1
                            _m2 = _ip2
                        else:
                            _m1 = _ip2
                            _m2 = _ip1
                        _s1, _s2, _s3 = _split_seg_2pts(_testseg, _m1, _m2)
                    elif _ip1 == _p1 or _ip1 == _p2:
                        if _ip2 != _p1 and _ip2 != _p2:
                            _s1, _s2 = split_segment(_testseg, _ip2)
                    elif _ip2 == _p1 or _ip2 == _p2:
                        if _ip1 != _p1 and _ip1 != _p2:
                            _s1, _s2 = split_segment(_testseg, _ip1)
                else:
                    raise ValueError, "Unexpected seg/circle count: %d" % _count
                #
                # if _s1 is not None then the test segment was split
                #
                if _s1 is not None:
                    _objstate[_testid] = False
                    bisect.insort_right(_seglist, _s1)
                    _objstate[id(_s1)] = True
                    _newsegs.append(_s1)
                if _s2 is not None:
                    bisect.insort_right(_seglist, _s2)
                    _objstate[id(_s2)] = True
                    _newsegs.append(_s2)
                if _s3 is not None:
                    bisect.insort_right(_seglist, _s3)
                    _objstate[id(_s3)] = True
                    _newsegs.append(_s3)
                #
                # if _arc1 is not None the the circle was split
                #
                if _arc1 is not None:
                    _objstate[_circid] = False
                    _newarcs.append(_arc1)
                if _arc2 is not None:
                    _newarcs.append(_arc2)
                break
    if len(_newarcs):
        _arclist = objdict['arcs']
        for _arc in _newarcs:
            layer.addObject(_arc)
            _arclist.append(_arc)
    _layercirclelist = []
    for _circ in _oldcircs:
        if _objstate[id(_circ)]:
            _layercirclelist.append(_circ)            
    objdict['circles'] = _layercirclelist
    _layerseglist = []
    for _seg in _newsegs:
        if _objstate[id(_seg)]:
            layer.addObject(_seg)
            _layerseglist.append(_seg)
        else:
            _p1, _p2 = _seg.getEndpoints()
            _p1.freeUser(_seg)
            _p2.freeUser(_seg)
    for _seg in _oldsegs:
        if _objstate[id(_seg)]:
            _layerseglist.append(_seg)
    objdict['segments'] = _layerseglist
    for _circ in _oldcircs:
        if not _objstate[id(_circ)]:
            layer.delObject(_circ)
    for _seg in _oldsegs:
        if not _objstate[id(_seg)]:
            layer.delObject(_seg)
 
def _seg_seg_split(layer, objdict):
    _seglist = objdict['segments']
    _seglist.sort()
    _segstate = {}
    _newsegs = []
    _oldsegs = []
    _objrects = objdict['objrects']
    for _seg in _seglist:
        _segstate[id(_seg)] = True
        _oldsegs.append(_seg)
    while len(_seglist):
        _testseg = _seglist.pop(0)
        _testid = id(_testseg)
        if _testid in _segstate and not _segstate[_testid]:
            continue
        if _testid not in _segstate:
            _segstate[_testid] = True
        _p1, _p2 = _testseg.getEndpoints()
        if _testid not in _objrects:
            _objrects[_testid] = _segment_bounds(_testseg)
        for _seg in _seglist:
            _segid = id(_seg)
            if (_segid in _segstate and not _segstate[_segid]):
                continue
            if _segid not in _segstate:
                _segstate[_segid] = True
            if _segid not in _objrects:
                _objrects[_segid] = _segment_bounds(_seg)
            if _no_overlap(_objrects[_testid], _objrects[_segid]):
                continue
            _p3, _p4 = _seg.getEndpoints()
            if _p1 == _p3 or _p1 == _p4 or _p2 == _p3 or _p2 == _p4:
                continue
            _ipts = intersections.find_intersections(_testseg, _seg)
            if len(_ipts):
                assert len(_ipts) == 1, "Invalid length: %d" % len(_ipts)
                _x, _y = _ipts[0]
                _ip = layer.find('point', _x, _y)
                if _ip is None:
                    _ip = Generic.point.Point(_x, _y)
                    layer.addObject(_ip)
                if _ip != _p1 and _ip != _p2:
                    _s1, _s2 = split_segment(_testseg, _ip)
                    _newsegs.append(_s1)
                    _segstate[id(_s1)] = True
                    _newsegs.append(_s2)
                    _segstate[id(_s2)] = True
                    bisect.insort_right(_seglist, _s1)
                    bisect.insort_right(_seglist, _s2)
                    _segstate[_testid] = False
                if _ip != _p3 and _ip != _p4:
                    _s1, _s2 = split_segment(_seg, _ip)
                    _newsegs.append(_s1)
                    _segstate[id(_s1)] = True
                    _newsegs.append(_s2)
                    _segstate[id(_s2)] = True
                    bisect.insort_right(_seglist, _s1)
                    bisect.insort_right(_seglist, _s2)
                    _segstate[_segid] = False
                break
    _layerseglist = []
    for _seg in _newsegs:
        if _segstate[id(_seg)]:
            layer.addObject(_seg)
            _layerseglist.append(_seg)
        else:
            _p1, _p2 = _seg.getEndpoints()
            _p1.freeUser(_seg)
            _p2.freeUser(_seg)
    for _seg in _oldsegs:
        if _segstate[id(_seg)]:
            _layerseglist.append(_seg)            
    objdict['segments'] = _layerseglist
    for _seg in _oldsegs:
        if not _segstate[id(_seg)]:
            layer.delObject(_seg)

def split_objects(layer, objlist):
    """Split a list of objects at their intersection points.

split_objects(layer, objlist):

layer: The layer containing the objects
objlist: The list of objects to split. This argument must be a list.
    """
    if not isinstance(layer, Generic.layer.Layer):
        raise TypeError, "Invalid layer: " + str(layer)
    if not isinstance(objlist, list):
        raise TypeError, "Invalid object list: " + str(objlist)
    _objdict = {}
    _segments = []
    _objdict['segments'] = _segments
    _arcs = []
    _objdict['arcs'] = _arcs
    _circles = []
    _objdict['circles'] = _circles
    _polylines = []
    _objdict['polylines'] = _polylines
    _objrects = {}
    _objdict['objrects'] = _objrects
    for _obj in objlist:
        _lobj = layer.findObject(_obj)
        if _lobj is not _obj:
            raise ValueError, "Object not found in layer: " + str(_obj)
        if isinstance(_obj, Generic.segment.Segment):
            _segments.append(_obj)
        elif isinstance(_obj, Generic.arc.Arc):
            _arcs.append(_obj)
        elif isinstance(_obj, Generic.circle.Circle):
            _circles.append(_obj)
        elif isinstance(_obj, Generic.polyline.Polyline):
            _polylines.append(_obj)
        else:
            pass
    if len(_objdict['segments']):
        _seg_seg_split(layer, _objdict)
        #
        # splitting circles produces arcs, so let the arc test
        # below handling segment/arc intersections ...
        #
        if len(_objdict['circles']):
            _seg_circle_split(layer, _objdict)
        if len(_objdict['polylines']):
            _seg_polyline_split(layer, _objdict)
    if len(_objdict['circles']):
        _circ_circ_split(layer, _objdict)
        if len(_objdict['arcs']):
            _circ_arc_split(layer, _objdict)
        if len(_objdict['polylines']):
            _circ_polyline_split(layer, _objdict)
    if len(_objdict['arcs']):
        if len(_objdict['segments']):
            _seg_arc_split(layer, _objdict)
        if len(_objdict['polylines']):
            _arc_polyline_split(layer, _objdict)
        _arc_arc_split(layer, _objdict)
    if len(_objdict['polylines']):
        _poly_poly_split(layer, _objdict)



#
# Copyright (c) 2002-2004 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Application Controller Object (& app delegate)
#

from math import hypot, pi, atan2

import objc

import Generic.tools
import CocoaConobjs
import CocoaEntities
import CocoaModify
import CocoaText
import CocoaDimensions
from PyObjCTools import NibClassBuilder
from Foundation import *
from AppKit import NSDocumentController, NSSavePanel, NSOKButton, NSCancelButton, NSPanel, NSWindowController, NSApplication, NSTextView

#
# Define True & False if this is Python < 2.2.1
#
try:
    True, False
except NameError:
    True, False = 1, 0
    
ZOOM = 2.0
    

NibClassBuilder.extractClasses("MainMenu")

class AppController(NibClassBuilder.AutoBaseClass):
    """ PythonCad application menu controller
    
    
Passes menu commands to the appropriate handlers
    """

#
# File menu
#
    def saveLayerAs_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return  # this is odd
        _sp = NSSavePanel.savePanel()
        _sp.setTitle_("Save Layer")
        _sp.setRequiredFileType_("xml.gz")
        _layer = _doc.getImage().getActiveLayer()
        _win = _doc.windowForSheet()
        _name = _layer.getName()
        _sp.beginSheetForDirectory_file_modalForWindow_modalDelegate_didEndSelector_contextInfo_(None, _name, _win, self, "savePanelDidEnd:returnCode:contextInfo:", 0)
        
    def savePanelDidEnd_returnCode_contextInfo_(self, sp, code, info):
        if NSOKButton == code:
            print "wahoo, we're saving the layer!"
        elif NSCancelButton == code:
            print "funk that - don't save the layer"
            
         
    savePanelDidEnd_returnCode_contextInfo_ = objc.selector(savePanelDidEnd_returnCode_contextInfo_, signature="v@:@ii")

#
# Edit menu commands
#
    def undo_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _doc.getImage().undo()
        _doc.getDA().setNeedsDisplay_(True)
                        
    def redo_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _doc.getImage().redo()    
        _doc.getDA().setNeedsDisplay_(True)
#
# Draw Basic Menu commands
#
    def drawPoint_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.PointTool()
        _doc.setTool(_tool)
        CocoaEntities.point_mode_init(_doc, _tool)
            
    def drawRect_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.RectangleTool()
        _doc.setTool(_tool)
        CocoaEntities.rectangle_mode_init(_doc, _tool)
    
    def drawSegment_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.SegmentTool()
        _doc.setTool(_tool)
        CocoaEntities.segment_mode_init(_doc, _tool)
    
    def drawCircleCentered_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.CircleTool()
        _doc.setTool(_tool)
        CocoaEntities.circle_center_mode_init(_doc, _tool)
    
    def drawCircle2Point_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.TwoPointCircleTool()
        _doc.setTool(_tool)
        CocoaEntities.circle_tp_mode_init(_doc, _tool)

    def drawArc_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.ArcTool()
        _doc.setTool(_tool)
        CocoaEntities.arc_center_mode_init(_doc, _tool)

#
# Draw Construction Line commands
#
    def drawHCLine_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.HCLineTool()
        _doc.setTool(_tool)
        CocoaConobjs.hcline_mode_init(_doc, _tool)

    def drawVCLine_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.VCLineTool()
        _doc.setTool(_tool)
        CocoaConobjs.vcline_mode_init(_doc, _tool)

    def drawACLine_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.ACLineTool()
        _doc.setTool(_tool)
        CocoaConobjs.acline_mode_init(_doc, _tool)
    
    def drawParallelCLine_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.ParallelOffsetTool()
        _doc.setTool(_tool)
        CocoaConobjs.cline_par_mode_init(_doc, _tool)

    def drawPerpendicularCLine_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.Tool()
        _doc.setTool(_tool)
        CocoaConobjs.cline_perp_mode_init(_doc, _tool)

    def drawTanCLine_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.Tool()
        _doc.setTool(_tool)
        CocoaConobjs.cline_tan_mode_init(_doc, _tool)
    
    def drawTan2CircleCLine_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.CCircleTangentLineTool()
        _doc.setTool(_tool)
        CocoaConobjs.cline_tan_2circ_mode_init(_doc, _tool)

    def draw2PointCLine_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.CLineTool()
        _doc.setTool(_tool)
        CocoaConobjs.cline_tp_mode_init(_doc, _tool)

#
# Draw construction circle commands
#
    def drawCCircleCenter_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.CCircleTool()
        _doc.setTool(_tool)
        CocoaConobjs.ccircle_center_mode_init(_doc, _tool)
    
    def drawCCircle2Point_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.TwoPointCCircleTool()
        _doc.setTool(_tool)
        CocoaConobjs.ccircle_tp_mode_init(_doc, _tool)

    def drawCCircle1Tan_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.TangentCCircleTool()
        _doc.setTool(_tool)
        CocoaConobjs.ccircle_tan1_mode_init(_doc, _tool)

    def drawCCircle2Tan_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.TwoPointTangentCCircleTool()
        _doc.setTool(_tool)
        CocoaConobjs.ccircle_tan2_mode_init(_doc, _tool)
        
#
# Draw more complicated things commands
#
    def drawChamfer_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.Tool()
        _doc.setTool(_tool)
        CocoaEntities.chamfer_mode_init(_doc, _tool)

    def drawFillet_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.Tool()
        _doc.setTool(_tool)
        CocoaEntities.fillet_mode_init(_doc, _tool)

    def drawLeader_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.LeaderTool()
        _doc.setTool(_tool)
        CocoaEntities.leader_mode_init(_doc, _tool)

    def drawPolyline_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.PolylineTool()
        _doc.setTool(_tool)
        CocoaEntities.polyline_mode_init(_doc, _tool)

    def drawPolygon_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.PolygonTool()
        _doc.setTool(_tool)
        self.openPolygonPanel()
        CocoaEntities.polygon_mode_init(_doc, _tool)

    def drawPolygonExternal_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.PolygonTool()
        _tool.setExternal()
        _doc.setTool(_tool)
        self.openPolygonPanel()
        CocoaEntities.polygon_mode_init(_doc, _tool)

    def drawText_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _win = _doc.windowForSheet()
        _app = NSApplication.sharedApplication()
        _app.beginSheet_modalForWindow_modalDelegate_didEndSelector_contextInfo_(self.textSheet, _win, None, None, 0)
        _tool = Generic.tools.TextTool()
        _doc.setTool(_tool)
        _textview = self.textSheet.initialFirstResponder()
        if isinstance(_textview, NSTextView):
            CocoaText.textview_format_setup(_doc, _textview)
        
    def drawTextSheetOK_(self, sender):
        _win = sender.window()
        _app = NSApplication.sharedApplication()
        _app.endSheet_(_win)
        _win.orderOut_(_win)
        _textview = _win.initialFirstResponder()
        if not isinstance(_textview, NSTextView):
            return
        _text = _textview.string()
        if not len(_text) > 0:
            return
        _doc = self.getCurrentDocument()
        _tool = _doc.getTool()
        if not isinstance(_tool, Generic.tools.TextTool):
            return
        _tool.setText(_text)
        CocoaText.text_entered(_doc, _tool, _textview)
    
    def drawTextSheetCancel_(self, sender):
        _win = sender.window()
        _app = NSApplication.sharedApplication()
        _app.endSheet_(_win)
        _win.orderOut_(_win)
        _textview = _win.initialFirstResponder()
        if not isinstance(_textview, NSTextView):
            return
        _textview.setString_("")

#
# Format menu 
#

#
# Modify menu
#
    def modifyMoveHorizontal_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.HorizontalMoveTool()
        _doc.setTool(_tool)
        CocoaModify.move_horizontal_init(_doc, _tool)

    def modifyMoveVertical_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.VerticalMoveTool()
        _doc.setTool(_tool)
        CocoaModify.move_vertical_init(_doc, _tool)

    def modifyMoveFree_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.MoveTool()
        _doc.setTool(_tool)
        CocoaModify.move_free_init(_doc, _tool)

    def modifyStretchHorizontal_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.HorizontalStretchTool()
        _doc.setTool(_tool)
        CocoaModify.stretch_horizontal_init(_doc, _tool)

    def modifyStretchVertical_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.VerticalStretchTool()
        _doc.setTool(_tool)
        CocoaModify.stretch_vertical_init(_doc, _tool)

    def modifyStretchFree_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.StretchTool()
        _doc.setTool(_tool)
        CocoaModify.stretch_free_init(_doc, _tool)

        
    def modifySplit_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.SplitTool()
        _doc.setTool(_tool)
        CocoaModify.split_init(_doc, _tool)
    
    def modifyMirror_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.MirrorTool()
        _doc.setTool(_tool)
        CocoaModify.mirror_init(_doc, _tool)
    
    def modifyTransfer_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.Tool()
        _doc.setTool(_tool)
        CocoaModify.transfer_init(_doc, _tool)

#
# Dimensions menu
#
    def dimensionLinear_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.LinearDimensionTool()
        _doc.setTool(_tool)
        CocoaDimensions.ldim_mode_init(_doc, _tool)
    
    def dimensionHorizontal_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.HorizontalDimensionTool()
        _doc.setTool(_tool)
        CocoaDimensions.ldim_mode_init(_doc, _tool)
   
    def dimensionVertical_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.VerticalDimensionTool()
        _doc.setTool(_tool)
        CocoaDimensions.ldim_mode_init(_doc, _tool)

    def dimensionRadial_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.RadialDimensionTool()
        _doc.setTool(_tool)
        CocoaDimensions.radial_mode_init(_doc, _tool)

    def dimensionAngular_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.AngularDimensionTool()
        _doc.setTool(_tool)
        CocoaDimensions.angular_mode_init(_doc, _tool)
#
# Window menu
#
    def windowZoom_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _tool = Generic.tools.RectangleTool()
        _doc.setTool(_tool)
        CocoaModify.zoom_mode_init(_doc, _tool)
        
    def windowZoomIn_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _da = _doc.getDA()
        _visRect = _da.visibleRect()
        _xinset = NSWidth(_visRect) / (ZOOM*2)
        _yinset = NSHeight(_visRect) / (ZOOM*2)
        _zoomRect = NSInsetRect(_visRect, _xinset, _yinset)
        _da.zoomToRect(_zoomRect)
        
    def windowZoomOut_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _da = _doc.getDA()
        _visRect = _da.visibleRect()
        _xinset = NSWidth(_visRect) / ZOOM
        _yinset = NSHeight(_visRect) / ZOOM
        _zoomRect = NSInsetRect(_visRect, -_xinset, -_yinset)
        _da.zoomToRect(_zoomRect)
    
    def windowZoomFit_(self, sender):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return
        _da = _doc.getDA()
        _da.fitImage()

#
# etc
#
    __polyPanelWindowController = None
    def openPolygonPanel(self):
        if self.__polyPanelWindowController is None:
            _wc = NSWindowController.alloc().initWithWindow_(self.polygonPanel)
            self.__polyPanelWindowController = _wc
        self.__polyPanelWindowController.showWindow_(self)
        
        
    def windowDidResignKey_(self, note):
        _win = note.object()
        if _win is self.polygonPanel:
            _doc = self.getCurrentDocument()
            if _doc is None:
                return
            _tool = _doc.getTool()
            if isinstance(_tool, Generic.tools.PolygonTool):
                _textField = _win.initialFirstResponder()
                _int = _textField.intValue()
                _tool.setSideCount(_int)
            
        
    def getCurrentDocument(self):
        _dc = NSDocumentController.sharedDocumentController()
        _doc = _dc.currentDocument()
        return _doc

    def validateMenuItem_(self, menuItem):
        _doc = self.getCurrentDocument()
        if _doc is None:
            return False
        _action = menuItem.action()
        _image = _doc.getImage()
        _layer = _image.getActiveLayer()
        # 
        # checks
        #
        if (_action == "undo:"):
            return _image.canUndo()
        
        elif (_action == "redo:"):
            return _image.canRedo()
        
        elif ((_action == "drawParallelCLine:") or
              (_action == "modifyMirror:")):
            _objs = _layer.getChildren()
            for _obj in _objs:
                if isinstance(_obj, (Generic.hcline.HCLine,
                                     Generic.vcline.VCLine,
                                     Generic.acline.ACLine,
                                     Generic.cline.CLine)):
                    return True
            return False
            
        elif (_action == "drawPerpendicularCLine:"):
            return _layer.hasChildren()
            
        elif (_action == "drawTanCLine:"):
            _circles = (_layer.getLayerEntities("circle") +
                        _layer.getLayerEntities("ccircle") +
                        _layer.getLayerEntities("arc"))
            if len(_circles) == 0:
                return False
            else:
                return True
                
        elif (_action == "drawTan2CircleCLine:"):
            _circles = _layer.getLayerEntities("ccircle")
            if len(_circles) < 2:
                return False
            return True
        
        elif (_action == "drawCCircle1Tan:"):
            _objs = _layer.getChildren()
            for _obj in _objs:
                if isinstance(_obj, Generic.conobject.ConstructionObject):
                    return True
            return False
        
        elif (_action == "drawCCircle2Tan:"):
            _ccircles = len(_layer.getLayerEntities("ccircle"))
            _cobjs = len(_layer.getLayerEntities("hcline") +
                      _layer.getLayerEntities("vcline") +
                      _layer.getLayerEntities("acline") +
                      _layer.getLayerEntities("cline"))
            if _cobjs == 0:
                return False
            elif (_ccircles + _cobjs) < 2:
                return False
            return True
            
        elif (_action == "modifyTransfer:"):
            _top = _image.getTopLayer()
            if len(_top.getSublayers()):
                return True
            return False
            
        elif (_action == "modifySplit:"):
            _objs = _layer.getChildren()
            for _obj in _objs:
                if isinstance(_obj, (Generic.segment.Segment,
                                     Generic.arc.Arc,
                                     Generic.circle.Circle,
                                     Generic.polyline.Polyline)):
                    return True
            return False
        
        return True





#
# Copyright (c) 2002, Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# handle setting and retrieving of the various options
# and preferences in an image
#
# none of these functions should be called directly - they
# are all meant as private functions for an Image object

import Generic.units
import Generic.dimension
import Generic.text
import Generic.style
import Generic.color
import Generic.linetype

#
# generic type tests
#

def _test_boolean_value(key, value):
    if not (value is True or
            value is False or
            value == 0 and
            value == 1):
        raise ValueError, "Invalid boolean for '%s'" + key

def _test_int_value(key, value, min=None, max=None):
    if not isinstance(value, int):
        raise TypeError, "Integer required for '%s'" % key
    if min is not None:
        assert isinstance(min, int), "Invalid minimum value: " + str(min)
        if value < min:
            raise ValueError, "Invalid value: %d < %d (min)" % (value, min)
    if max is not None:
        assert isinstance(max, int), "Invalid maximum value: " + str(max)
        if value > max:
            raise ValueError, "Invalid value: %d > %d (max)" % (value, max)

def _test_float_value(key, value, min=None, max=None):
    if not isinstance(value, float):
        raise TypeError, "Float required for '%s'" % key
    if min is not None:
        assert isinstance(min, float), "Invalid minimum value: " + str(min)
        if value < min:
            raise ValueError, "Invalid value: %g < %g (min)" % (value, min)
    if max is not None:
        assert isinstance(max, float), "Invalid maximum value: " + str(max)
        if value > max:
            raise ValueError, "Invalid value: %g > %g (max)" % (value, max)

def _test_unicode_value(key, value):
    if not isinstance(value, unicode):
        raise TypeError, "Unicode string required for '%s'" % key

def _test_color_value(key, value):
    if not isinstance(value, Generic.color.Color):
        raise TypeError, "Color object required for '%s'" % key

def _test_linetype_value(key, value):
    if not isinstance(value, Generic.linetype.Linetype):
        raise TypeError, "Invalid line type for '%s'" % key

def _test_linestyle_value(key, value):
    if not isinstance(value, Generic.style.Style):
        raise TypeError, "Invalid text style for '%s'" % key
    
def _test_font_family(key, value):
    if not isinstance(value, str):
        raise TypeError, "String required for '%s'" % key
    if value == '':
        raise ValueError, "Non-null string required for '%s'" % key

def _test_dimstyle(key, value):
    if not isinstance(value, Generic.dimension.DimStyle):
        raise TypeError, "Invalid DimStyle for '%s'" % key
    
def _test_font_weight(key, value):
    if (value != Generic.text.NORMAL and
        value != Generic.text.LIGHT and
        value != Generic.text.BOLD and
        value != Generic.text.HEAVY):
        raise ValueError, "Invalid font weight for '%s'" % key

def _test_font_style(key, value):
    if (value != Generic.text.NORMAL and
        value != Generic.text.OBLIQUE and
        value != Generic.text.ITALIC):
        raise ValueError, "Invalid font style for '%s'" % key

def _test_units(key, value):
    if (value != Generic.units.MILLIMETERS and
        value != Generic.units.MICROMETERS and
        value != Generic.units.METERS and
        value != Generic.units.KILOMETERS and
        value != Generic.units.INCHES and
        value != Generic.units.FEET and
        value != Generic.units.YARDS and
        value != Generic.units.MILES):
        raise ValueError, "Invalid unit for '%s'" %  key

#
# dimension related tests
#

def _set_dim_style(opt, ds):
    _test_dimstyle(opt, ds)
    
def _set_dim_primary_font_family(opt, family):
    _test_font_family(opt, family)

def _set_dim_primary_font_size(opt, size):
    _test_int_value(opt, size, min=1)

def _set_dim_primary_font_style(opt, st):
    _test_font_style(opt, st)

def _set_dim_primary_font_weight(opt, weight):
    _test_font_weight(opt, weight)

def _set_dim_primary_font_color(opt, col):
    _test_color_value(opt, col)

def _set_dim_primary_prefix(opt, pfx):
    _test_unicode_value(opt, pfx)

def _set_dim_primary_suffix(opt, pfx):
    _test_unicode_value(opt, pfx)

def _set_dim_primary_units(opt, u):
    _test_units(opt, u)

def _set_dim_primary_precision(opt, prec):
    _test_int_value(opt, prec, min=0, max=15)

def _set_dim_primary_print_zero(opt, flag):
    _test_boolean_value(opt, flag)

def _set_dim_primary_trail_decimal(opt, flag):
    _test_boolean_value(opt, flag)

def _set_dim_secondary_font_family(opt, family):
    _test_font_family(opt, family)

def _set_dim_secondary_font_size(opt, size):
    _test_int_value(opt, size, min=1)

def _set_dim_secondary_font_style(opt, st):
    _test_font_style(opt, st)

def _set_dim_secondary_font_weight(opt, weight):
    _test_font_weight(opt, weight)

def _set_dim_secondary_font_color(opt, col):
    _test_color_value(opt, col)

def _set_dim_secondary_prefix(opt, pfx):
    _test_unicode_value(opt, pfx)

def _set_dim_secondary_suffix(opt, pfx):
    _test_unicode_value(opt, pfx)

def _set_dim_secondary_units(opt, u):
    _test_units(opt, u)

def _set_dim_secondary_precision(opt, prec):
    _test_int_value(opt, prec, min=0, max=15)

def _set_dim_secondary_print_zero(opt, flag):
    _test_boolean_value(opt, flag)

def _set_dim_secondary_trail_decimal(opt, flag):
    _test_boolean_value(opt, flag)

def _set_dim_offset(opt, offset):
    _test_float_value(opt, offset, min=0.0)

def _set_dim_extension(opt, ext):
    _test_float_value(opt, ext, min=0.0)

def _set_dim_color(opt, col):
    _test_color_value(opt, col)

def _set_dim_dual_mode(opt, mode):
    _test_boolean_value(opt, mode)

def _set_dim_position(opt, pos):
    if (pos != Generic.dimension.Dimension.SPLIT and
        pos != Generic.dimension.Dimension.ABOVE and
        pos != Generic.dimension.Dimension.BELOW):
        raise ValueError, "Invalid dimension position: " + str(pos)

def _set_dim_endpoint(opt, ept):
    if (ept != Generic.dimension.Dimension.NO_ENDPOINT and
        ept != Generic.dimension.Dimension.ARROW and
        ept != Generic.dimension.Dimension.FILLED_ARROW and
        ept != Generic.dimension.Dimension.SLASH and
        ept != Generic.dimension.Dimension.CIRCLE):
        raise ValueError, "Invalid endpoint value: " + str(ept)

def _set_dim_endpoint_size(opt, size):
    _test_float_value(opt, size, min=0.0)

def _set_radial_dim_primary_prefix(opt, pfx):
    _test_unicode_value(opt, pfx)

def _set_radial_dim_primary_suffix(opt, sfx):
    _test_unicode_value(opt, sfx)

def _set_radial_dim_secondary_prefix(opt, sfx):
    _test_unicode_value(opt, sfx)

def _set_radial_dim_secondary_suffix(opt, pfx):
    _test_unicode_value(opt, pfx)

def _set_radial_dim_dia_mode(opt, flag):
    _test_boolean_value(opt, flag)

def _set_angular_dim_primary_prefix(opt, pfx):
    _test_unicode_value(opt, pfx)

def _set_angular_dim_primary_suffix(opt, sfx):
    _test_unicode_value(opt, sfx)

def _set_angular_dim_secondary_prefix(opt, pfx):
    _test_unicode_value(opt, pfx)

def _set_angular_dim_secondary_suffix(opt, sfx):
    _test_unicode_value(opt, sfx)

def _set_angular_dim_small_angle_mode(opt, flag):
    _test_boolean_value(opt, flag)

#
# text related tests
#

def _set_text_style(opt, textstyle):
    if not isinstance(textstyle, Generic.text.TextStyle):
        raise TypeError, "Invalid text style: " + `textstyle`

def _set_font_family(opt, family):
    _test_font_family(opt, family)

def _set_font_style(opt, st):
    _test_font_style(opt, st)

def _set_font_weight(opt, weight):
    _test_font_weight(opt, weight)

def _set_font_size(opt, size):
    _test_int_value(opt, size, min=1)

def _set_font_color(opt, col):
    _test_color_value(opt, col)

def _set_chamfer_length(opt, length):
    _test_float_value(opt, length, min=0.0)

def _set_fillet_radius(opt, radius):
    _test_float_value(opt, radius)

def _set_line_style(opt, st):
    _test_linestyle_value(opt, st)

def _set_line_color(opt, col):
    _test_color_value(opt, col)

def _set_line_type(opt, lt):
    _test_linetype_value(opt, lt)

def _set_line_thickness(opt, thickness):
    _test_float_value(opt, thickness, min=0.0)

def _set_units(opt, u):
    _test_units(opt, u)

def _set_highlight_points(opt, flag):
    _test_boolean_value(opt, flag)

def _set_inactive_layer_color(opt, col):
    _test_color_value(opt, col)

def _set_background_color(opt, col):
    _test_color_value(opt, col)

def _set_single_point_color(opt, col):
    _test_color_value(opt, col)

def _set_multi_point_color(opt, col):
    _test_color_value(opt, col)
    
def _set_autosplit(opt, col):
    _test_boolean_value(opt, col)

def _leader_arrow_size(opt, size):
    _test_float_value(opt, size, min=0.0)

#
# the keys of this dictionary represent options that have
# tests to validate the option value is suitable
#

_optdict = {
    'DIM_STYLE' : _set_dim_style,
    'DIM_PRIMARY_FONT_FAMILY' : _set_dim_primary_font_family,
    'DIM_PRIMARY_FONT_SIZE' : _set_dim_primary_font_size,
    'DIM_PRIMARY_FONT_WEIGHT' : _set_dim_primary_font_weight,
    'DIM_PRIMARY_FONT_STYLE' : _set_dim_primary_font_style,
    'DIM_PRIMARY_FONT_COLOR' : _set_dim_primary_font_color,
    'DIM_PRIMARY_PREFIX' : _set_dim_primary_prefix,
    'DIM_PRIMARY_SUFFIX' : _set_dim_primary_suffix,
    'DIM_PRIMARY_PRECISION' : _set_dim_primary_precision,
    'DIM_PRIMARY_UNITS' : _set_dim_primary_units,
    'DIM_PRIMARY_LEADING_ZERO' : _set_dim_primary_print_zero,
    'DIM_PRIMARY_TRAILING_DECIMAL': _set_dim_primary_trail_decimal,
    'DIM_SECONDARY_FONT_FAMILY' : _set_dim_secondary_font_family,
    'DIM_SECONDARY_FONT_SIZE' : _set_dim_secondary_font_size,
    'DIM_SECONDARY_FONT_WEIGHT' : _set_dim_secondary_font_weight,
    'DIM_SECONDARY_FONT_STYLE' : _set_dim_secondary_font_style,
    'DIM_SECONDARY_FONT_COLOR' : _set_dim_secondary_font_color,
    'DIM_SECONDARY_PREFIX' : _set_dim_secondary_prefix,
    'DIM_SECONDARY_SUFFIX' : _set_dim_secondary_suffix,
    'DIM_SECONDARY_PRECISION' : _set_dim_secondary_precision,
    'DIM_SECONDARY_UNITS' : _set_dim_secondary_units,
    'DIM_SECONDARY_LEADING_ZERO' : _set_dim_secondary_print_zero,
    'DIM_SECONDARY_TRAILING_DECIMAL': _set_dim_secondary_trail_decimal,
    'DIM_OFFSET' : _set_dim_offset,
    'DIM_EXTENSION': _set_dim_extension,
    'DIM_COLOR' : _set_dim_color,
    'DIM_POSITION': _set_dim_position,
    'DIM_ENDPOINT': _set_dim_endpoint,
    'DIM_ENDPOINT_SIZE' : _set_dim_endpoint_size,
    'DIM_DUAL_MODE' : _set_dim_dual_mode,
    'RADIAL_DIM_PRIMARY_PREFIX' : _set_radial_dim_primary_prefix,
    'RADIAL_DIM_PRIMARY_SUFFIX' : _set_radial_dim_primary_suffix,
    'RADIAL_DIM_SECONDARY_PREFIX' : _set_radial_dim_secondary_prefix,
    'RADIAL_DIM_SECONDARY_SUFFIX' : _set_radial_dim_secondary_suffix,
    'RADIAL_DIM_DIA_MODE' : _set_radial_dim_dia_mode,
    'ANGULAR_DIM_PRIMARY_PREFIX' : _set_angular_dim_primary_prefix,
    'ANGULAR_DIM_PRIMARY_SUFFIX' : _set_angular_dim_primary_suffix,
    'ANGULAR_DIM_SECONDARY_PREFIX' : _set_angular_dim_secondary_prefix,
    'ANGULAR_DIM_SECONDARY_SUFFIX' : _set_angular_dim_secondary_suffix,
    'TEXT_STYLE' : _set_text_style,
    'FONT_FAMILY' : _set_font_family,
    'FONT_STYLE' : _set_font_style,
    'FONT_WEIGHT' : _set_font_weight,
    'FONT_SIZE' : _set_font_size,
    'FONT_COLOR' : _set_font_color,
    'CHAMFER_LENGTH' : _set_chamfer_length,
    'FILLET_RADIUS' : _set_fillet_radius,
    'UNITS' : _set_units,
    'LINE_STYLE' : _set_line_style,
    'LINE_COLOR' : _set_line_color,
    'LINE_TYPE' : _set_line_type,
    'LINE_THICKNESS': _set_line_thickness,
    'HIGHLIGHT_POINTS' : _set_highlight_points,
    'INACTIVE_LAYER_COLOR' : _set_inactive_layer_color,
    'BACKGROUND_COLOR' : _set_background_color,
    'SINGLE_POINT_COLOR' : _set_single_point_color,
    'MULTI_POINT_COLOR' : _set_multi_point_color,
    'AUTOSPLIT' : _set_autosplit,
    'LEADER_ARROW_SIZE' : _leader_arrow_size,
    }

#
# the test_option() function will return True for tested
# options that validate or for unknown options. Invalid
# tested options will raise an exception that the caller
# must handle
#

def test_option(opt, val):
    _valid = True
    if opt in _optdict:
        _optdict[opt](opt, val)
    return _valid



#
# Copyright (c) 2002, 2003 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# hatching code
#

import math
import point
import segment
import circle
import arc
import segjoint
import layer

class Path(object):
    """The class for maintaining a list of objects defining a hatch border.

A Path object contains one or more objects defining the boundary
of a hatching. If the Path length is 1, the boundary is either a
circle or an arc where the start angle and end angle are equal.
There is no upper limit to the number of objects in the path.

If a Path consists of Segments, Arcs, Chamfer, or Fillets, the
Path can only be valid if starting at any point in any object
in the Path the connections between the objects lead back to the
starting point.

A Path has the following methods:

isExternal(): Check if the Path is an outer boundary
isCircular(): Check if the Path is a Circle or closed Arc.
getPath(): Return the objects comprising the Path
inPath(): Test if a coordinate is inside a Path.
    """
    def __init__(self, objs, external=True):
        """Initialize a Path.

p = Path(objs[, external])

The required argument "objs" is a list of objects defining
the path. The valid objects are circles, arcs, segments,
chamfers and fillets.

The optional argument "external" is by default True, meaning
that this Path is an outer boundary. If the argument is False,
then the Path represents an internal non-hatched area inside
another Path.
        """
        if not isinstance(objs, list):
            raise TypeError, "Invalid object list: " + `objs`
        if not len(objs):
            raise ValueError, "Invalid empty object list"
        for _obj in objs:
            _valid = False
            if (isinstance(_obj, segment.Segment) or
                isinstance(_obj, arc.Arc) or
                isinstance(_obj, circle.Circle) or
                isinstance(_obj, segjoint.Chamfer) or
                isinstance(_obj, segjoint.Fillet)):
                _valid = True
            if not _valid:
                raise TypeError, "Invalid object in object list: " + `_obj`
        _circular = False
        if len(objs) == 1:
            _circular = True
            _obj = objs[0]
            _valid = False
            if isinstance(_obj, arc.Arc):
                _sa = _obj.getStartAngle()
                _ea = _obj.getEndAngle()
                if abs(_sa - _ea) < 1e-10:
                    _valid = True
                if not _valid:
                    raise TypeError, "Invalid single Arc path: " + str(_obj)
            elif isinstance(_obj, circle.Circle):
                pass
            else:
                raise TypeError, "Invalid single entity path: " + str(_obj)
        else:
            _valid = True
            for _obj in objs:
                if isinstance(_obj, circle.Circle):
                    if not isinstance(_obj, arc.Arc):
                        _valid = False
                        break
            if not _valid:
                raise TypeError, "Circle found in multi-object path"
            _valid = True # _validate_path(objlist)
            if not _valid:
                raise ValueError, "Objlist objects do not make a closed path."
        if external is not True and external is not False:
            raise ValueError, "Invalid external flag: " + `external`
        self.__objs = objs[:]
        self.__circular = _circular
        self.__external = external

    def __len__(self):
        return len(self.__objs)

    def __str__(self):
        if self.__external:
            print "External Path: ["
        else:
            print "Internal Path: ["
        for _obj in self.__objs:
            print str(_obj)
        print "]"

    def isExternal(self):
        """Test if the Path is an external border.

isExternal()
        """
        return self.__external

    def isCircular(self):
        """Test if the Path is a Circle or closed Arc.

isCircular()        
        """
    def getPath(self):
        """Return the objects defining the Path.

getPath()

This method returns a list of objects.
        """
        return self.__objs[:]

    def inPath(self, x, y):
        """Test if a coordinate pair are inside a Path.

inPath(x, y)

This method has two required arguments:

x: A float giving the "x" coordinate.
y: A float giving the "y" coordinate.

This method returns True if the Point is inside the
Path, and False otherwise.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _inside = False
        if self.__circular:
            assert len(self.__objs) == 1, "Invalid circular path length"
            _circ = self.__objs[0]
            _cx, _cy = _circ.getCenter().getCoords()
            _sep = math.hypot((_cx - _x), (_cy - _y))
            if _sep < _circ.getRadius():
                _inside = True
        else:
            _xp = 0.0
            _yp = 0.0
            _idx = None
            for _i in range(len(self.__objs)):
                _obj = self.__objs[_i]
                if isinstance(_obj, segment.Segment):
                    _p1, _p2 = _obj.getEndpoints()
                    _p1x, _p1y = _p1.getCoords()
                    _p2x, _p2y = _p2.getCoords()
                elif isinstance(_obj, arc.Arc):
                    _ep1, _ep2 = _obj.getEndpoints()
                    _p1x, _p1y = _ep1
                    _p2x, _p2y = _ep2
                elif isinstance(_obj, (segjoint.Chamfer, segjoint.Fillet)):
                    _p1, _p2 = _obj.getMovingPoints()
                    _p1x, _p1y = _p1.getCoords()
                    _p2x, _p2y = _p2.getCoords()
                _xdiff = _p2x - _p1x
                _ydiff = _p2y - _p1y
                _sqlen = pow(_xdiff, 2) + pow(_ydiff, 2)
                _r = ((_x - _p1x)*(_xdiff) + (_y - _p1y)*(_ydiff))/_sqlen
                if 0.0 < _r < 1.0:
                    _s = ((_p1y - _y)*(_xdiff) - (_p1x - _x)*(_ydiff))/_sqlen
                    if abs(_s) > 1e-10:
                        _xp = _p1x + (_r * _xdiff)
                        _yp = _p1y + (_r * _ydiff)
                        _idx = _i
                        break
            if _idx is not None:
                _count = 1
                for _i in range(len(self.__objs)):
                    if _i == _idx:
                        continue
                    _obj = self.__objs[_i]
                    if isinstance(_obj, segment.Segment):
                        _p1, _p2 = _obj.getEndpoints()
                        _p1x, _p1y = _p1.getCoords()
                        _p2x, _p2y = _p2.getCoords()
                    elif isinstance(_obj, arc.Arc):
                        _ep1, _ep2 = _obj.getEndpoints()
                        _p1x, _p1y = _ep1
                        _p2x, _p2y = _ep2
                    elif isinstance(_obj, (segjoint.Chamfer, segjoint.Fillet)):
                        _p1, _p2 = _obj.getMovingPoints()
                        _p1x, _p1y = _p1.getCoords()
                        _p2x, _p2y = _p2.getCoords()
                    _d = ((_p2x - _p1x)*(_yp - _y)) - ((_p2y - _p1y)*(_xp - _x))
                    if abs(_d) > 1e-10:
                        _n = ((_p1y - _y)*(_xp - _x)) - ((_p1x - _x)*(_yp - _y))
                        _r = _n/_d
                        if 0.0 < _r < 1.0:
                            _count = _count + 1
                if _count % 2: # need to test if point is in an arc ...
                    _inside = True
        return _inside
    
class HatchRegion(object):
    """The class defining a hatched area.

A HatchRegion object consists of one Path object defining
the external boundary of the hatching, and a list of zero
or more Paths defining any areas inside the enclosing Path
that are not hatched.
    """
    def __init__(self, extpath, voids=[]):
        """Initialize a HatchRegion.

h = HatchRegion(extpath[, voids])

The required argument "extpath" is a Path object defining
the external boundary of the hatching. The optional argument
"voids" is a list of Path objects defining areas within the
external Path that are not to be hatched.

A HatchRegion has the following methods:

        """
        if not isinstance(extpath, Path):
            raise TypeError, "Invalid external path: " + `extpath`
        if not extpath.isExternal():
            raise ValueError, "Path not defined to be an external path: " + `extpath`
        if not isinstance(voids, list):
            raise TypeError, "Invalid void list: " + `voids`
        for _void in voids:
            if not isinstance(_void, Path):
                raise TypeError, "Invalid path in void list: " + `_void`
            if _void.isExternal():
                raise ValueError, "Void area defined as external: " + `_void`
        self.__ext_path = extpath
        self.__voids = voids[:]

    def getExternalPath(self):
        """Return the external Path for the HatchRegion.

getExternalPath()        
        """
        return self.__ext_path

    def hasVoids(self):
        """Test if the HatchRegion has any internal non-hatched areas.

hasVoids()        
        """
        return len(self.__voids) > 0

    def getVoids(self):
        """Get any internal areas in the HatchRegion.

getVoids()        
        """
        return self.__voids[:]

def _seg_seg_touch(sega, segb):
    _touch = False
    _pa1, _pa2 = sega.getEndpoints()
    _pb1, _pb2 = segb.getEndpoints()
    if _pa1 is _pb1 or _pa1 is _pb2 or _pa2 is _pb1 or _pa2 is _pb2:
        _touch = True
    return _touch

def _seg_arc_touch(seg, a):
    _touch = False
    _p1, _p2 = seg.getEndpoints()
    _ep1, _ep2 = a.getEndpoints()
    if _p1 == _ep1 or _p1 == _ep2 or _p2 == _ep1 or _p2 == _ep2:
        _touch = True
    return _touch

def _arc_arc_touch(arca, arcb):
    _touch = False
    _aep1, _aep2 = arca.getEndpoints()
    _bep1, _bep2 = arcb.getEndpoints()
    if _aep1 == _bep1 or _aep1 == _bep2 or _aep2 == _bep1 or _aep2 == _bep2:
        _touch = True
    return _touch

def _seg_joint_touch(seg, joint):
    _touch = False
    _s1, _s2 = joint.getSegments()
    if _s1 is seg or _s2 is seg:
        _touch = True
    return _touch
    
def _old_validate_path(objlist):
    """Test if the objects in the objlist make a closed path.

_validate_path(objlist)

This function is private the the hatching code.    
    """
    if not isinstance(objlist, list):
        raise TypeError, "Invalid object list: " + `objlist`
    _startpt = None
    _nextpt = None
    _valid = False
    for _obj in objlist:
        print "testing object: " + `_obj`
        print "start: " + `_startpt`
        print "next: " + `_nextpt`
        if isinstance(_obj, segment.Segment):
            _p1, _p2 = _obj.getEndpoints()
            if _startpt is None:
                _startpt = _p1
                _nextpt = _p2
            else:
                if _nextpt == _p1:
                    _nextpt = _p2
                elif _nextpt == _p2:
                    _nextpt = _p1
                else:
                    break
        elif isinstance(_obj, arc.Arc):
            _ep1, _ep2 = _obj.getEndpoints()
            if _startpt is None:
                _startpt = _ep1
                _nextpt = _ep2
            else:
                if _nextpt == _ep1:
                    _nextpt = _ep2
                elif _startpt == _ep2:
                    _nextpt = _ep1
                else:
                    break
        elif isinstance(_obj, (segjoint.Chamfer, segjoint.Fillet)):
            _p1, _p2 = _obj.getMovingPoints()
            if _startpt is None:
                _startpt = _p1
                _nextpt = _p2
            else:
                if _nextpt == _p1:
                    _nextpt = _p2
                elif _nextpt == _p2:
                    _nextpt = _p1
                else:
                    break
        else:
            raise TypeError, "Invalid object in path: " + `_obj`
    if _startpt == _nextpt:
        _valid = True
    return _valid

def _can_touch(obja, objb):
    _touch = False
    if isinstance(obja, segment.Segment):
        if isinstance(objb, segment.Segment):
            _touch = _seg_seg_touch(obja, objb)
        elif isinstance(objb, arc.Arc):
            _touch = _seg_arc_touch(obja, objb)
        elif isinstance(objb, segjoint.SegJoint):
            _touch = _seg_joint_touch(obja, objb)
    elif isinstance(obja, arc.Arc):
        if isinstance(objb, segment.Segment):
            _touch = _seg_arc_touch(objb, obja)
        elif isinstance(objb, arc.Arc):
            _touch = _arc_arc_touch(obja, objb)
    elif isinstance(obja, segjoint.SegJoint):
        if isinstance(objb, segment.Segment):
            _touch = _seg_joint_touch(objb, obja)
    return _touch

def _validate_path(lyr, objlist):
    """Test if the objects in the objlist make a closed path.

_validate_path(objlist)

This function is private the the hatching code.    
    """
    if not isinstance(objlist, list):
        raise TypeError, "Invalid object list: " + `objlist`
    for _obj in objlist:
        _lobj = lyr.findObject(_obj)
        if _lobj is not _obj:
            raise ValueError, "Object not in layer: " + `_obj`
    _valid = True
    for _i in range(len(objlist) - 1):
        _obja = objlist[_i]
        _objb = objlist[_i+1]
        if not _can_touch(_obja, _objb):
            _valid = False
            break
    if _valid:
        _valid = _can_touch(objlist[0], objlist[-1])
    return _valid

def point_boundaries(plist):
    _xmin = None
    _xmax = None
    _ymin = None
    _ymax = None
    _set = False
    if len(plist) > 1:
        for _pt in plist:
            _x, _y = _pt.getCoords()
            if not _set:
                _xmin = _x
                _xmax = _x
                _ymin = _y
                _ymax = _y
            else:
                if _x < _xmin:
                    _xmin = _x
                if _x > _xmax:
                    _xmax = _x
                if _y < _ymin:
                    _ymin = _y
                if _y > _ymax:
                    _ymax = _y
    return (_xmin, _ymin, _xmax, _ymax)

def point_in_path(path):
    hits = 0
    for seg in path:
        p1, p2 = seg.getEndpoints()
        p1x, p1y = p1.getEndpoints()
        p2x, p2y = p2.getEndpoints()
        xmin = min(p1x, p2x)
        xmax = max(p1x, p2x)
        ymin = min(p1y, p2y)
        ymax = max(p1y, p2y)
        if hx < xmin or hx > max or hy > ymax:
            continue
        hits = 1 - hits
    return hits
        
def draw_path(path):
    if len(path):
        print "path: ["
        for seg in path:
            print seg
        print "]"

def make_paths(pt, seg, sdict):
    paths = []
    print "initial segment: " + str(seg)
    sp1, sp2 = seg.getEndpoints()
    if pt is sp1:
        sp = sp1
    else:
        sp = sp2
    print "start point: " + str(sp)
    segkeys = {}
    path = []
    segkeys[seg] = pt
    path.append(seg)
    draw_path(path)
    paths.append(path)
    while(len(paths)):
        path = paths.pop()
        draw_path(path)
        seg = path[-1]
        print "path final segment: " + str(seg)
        # print "segkey: " + str(segkeys[seg])
        p1, p2 = seg.getEndpoints()
        print "p1: " + str(p1)
        if p1 in sdict and segkeys[seg] is not p1:
            if p1 is sp:
                print "complete path:"
                draw_path(path)
            else:
                for p1seg in sdict[p1]:
                    if p1seg not in path:
                        segkeys[p1seg] = p1
                        path.append(p1seg)
                        # print "new path:"
                        draw_path(path)
                        paths.append(path)
        print "p2: " + str(p2)
        if p2 in sdict and segkeys[seg] is not p2:
            if p2 is sp:
                print "complete path:"
                draw_path(path)
            else:
                for p2seg in sdict[p2]:
                    if p2seg not in path:
                        segkeys[p2seg] = p2
                        path.append(p2seg)
                        # print "new_path:"
                        draw_path(path)
                        paths.append(path)


def check_clist(ct, clist):
    xct, yct = ct.getCenter().getCoords()
    rct = ct.getRadius()
    add_flag = True
    i = 0
    while (i < len(clist)):
        _c = clist[i]
        x, y = _c.getCenter().getCoords()
        r = _c.getRadius()
        sep = math.hypot((xct - x), (yct - y))
        if sep < r: # ct center point inside _c
            if sep + rct < r:
                add_flag = False
            else:
                i = i + 1
        elif sep < rct: # _c center point inside ct
            if sep + r < rct:
                del clist[i]
            else:
                i = i + 1
        else: # two circle that may or may not overlap
            i = i + 1
        if not add_flag:
            break
    return add_flag
            
def get_contained_circles(l, c):
    clist = []
    xc, yc = c.getCenter().getCoords()
    rc = c.getRadius()
    for _cir in l.getLayerEntities("circle"):
        if _cir is c:
            continue
        x, y = _cir.getCenter().getCoords()
        r = _cir.getCoords()
        sep = math.hypot((xc - x), (yc - y))
        if sep + r < rc:
            if(check_clist(_cir, clist)):
                clist.append(_cir)
    return clist

def make_hatch_area(lyr, x, y):
    _x = x
    if not isinstance(_x, float):
        _x = float(x)
    _y = y
    if not isinstance(_y, float):
        _y = float(y)
    if not isinstance(lyr, layer.Layer):
        raise TypeError, "Invalid layer: " + `lyr`
    #
    # see if we're in a circle
    #
    _circle = None
    for _c in lyr.getLayerEntities("circle"):
        _xc, _yc = _c.getCenter().getCoords()
        _r = _c.getRadius()
        _sep = math.hypot((_xc - _x), (_yc - _y))
        if _sep < _r:
            if _circle is None:
                _circle = _c
            else:
                _rc = _circle.getRadius()
                if _r < _rc:
                    _circle = _c
    #
    # get the eligible points in the layer and
    # store any circles that may be fully inside
    # the bounding circle
    #
    _pts = {}
    _circle_voids = []
    if _circle is not None:
        _cx, _cy = _circle.getCenter().getCoords()
        _rad = _circle.getRadius()
        _xmin = _cx - _rad
        _ymin = _cy - _rad
        _xmax = _cx + _rad
        _ymax = _cy + _rad
        for _pt in lyr.getLayerEntities("point"):
            _x, _y = _pt.getCoords()
            if _x < _xmin or _y < _ymin or _y > _ymax:
                continue
            if _x > _xmax:
                break
            _sep = math.hypot((_cx - _x), (_cy - _y))
            if _sep < _rad:
                _addpt = True
                if lyr.countUsers(_pt) == 1:
                    _obj = lyr.usesObject(_pt)
                    if not isinstance(_obj, circle.Circle):
                        _addpt = False
                if _addpt:
                    _pts[_pt] = True
        for _circ in lyr.getLayerEntities("circle"):
            if _circ is _circle:
                continue
            _tcx, _tcy = _circ.getCenter().getCoords()
            _tr = _circ.getRadius()
            if (_tcx + _tr) > _xmax:
                break
            if ((_tcx - _tr) < _xmin or
                (_tcy - _tr) < _ymin or
                (_tcy + _tr) > _ymax):
                continue
            _sepmax = math.hypot((_cx - _tcx), (_cy - _tcy)) + _tr
            if _sepmax < _rad:
                _circle_voids.append(_circ)
    else:
        for _pt in lyr.getLayerEntities("point"):
            _addpt = True
            if lyr.countUsers(_pt) == 1:
                _obj = lyr.usesObject(_pt)
                if not isinstance(_obj, circle.Circle):
                    _addpt = False
            if _addpt:
                _pts[_pt] = True
    #
    # find the entites that can make closed paths
    #
    _objs = {}
    for _pt in _pts:
        for _user in lyr.usesObject(_pt):
            if isinstance(_user, (segment.Segment, arc.Arc,
                                  segjoint.Chamfer, segjoint.Fillet)):
                _objs[_user] = True
                if isinstance(_user, segment.Segment):
                    for _seguser in lyr.usesObject(_user):
                        _objs[_seguser] = True
    _paths = {}
    for _obj in _objs:
        _p1 = None
        _p2 = None
        if isinstance(_obj, segment.Segment):
            if _obj not in _paths:
                _paths[_obj] = []
            for _user in lyr.usesObject(_obj):
                if isinstance(_user, (segjoint.Chamfer, segjoint.Fillet)):
                    _paths[_obj].append(_user)
            _p1, _p2 = _obj.getEndpoints()
        elif isinstance(_obj, arc.Arc):
            _ep1, _ep2 = _obj.getEndpoints()
            for _pt in _pts:
                if _pt == _ep1:
                    _p1 = _pt
                elif _p2 == _ep2:
                    _p2 = _pt
                if _p1 is not None and _p2 is not None:
                    break
            if _p1 is None or _p2 is None:
                continue # only one arc endpoint in list
        else:
            _s1, _s2 = _obj.getSegments()
            if _obj not in _paths:
                _paths[_obj] = []
            _paths[_obj].append(_s1)
            _paths[_obj].append(_s2)
        if _p1 is not None:
            for _user in lyr.usesObject(_p1):
                if _user is not _obj:
                    if isinstance(_user, (segment.Segment, arc.Arc)):
                        _paths[_obj].append(_user)
        if _p2 is not None:
            for _user in lyr.usesObject(_p2):
                if _user is not _obj:
                    if isinstance(_user, (segment.Segment, arc.Arc)):
                        _paths[_obj].append(_user)
    #
    # remove any object that doesn't connect to another object
    #
    _objlist = _paths.keys()
    for _obj in _objlist:
        if len(_paths[_obj]) < 1:
            del _paths[_obj]
    #
    # try to make paths from the selected objects
    #
    _routes = {}
    _objlist = _paths.keys()
    _objcount = len(_objlist)
    for _obj in _objlist:
        _objpaths = []
        _path = [_obj]
        for _fullpath in _make_paths(_paths, _objcount, _obj, _path):
            print "path: " + str(_fullpath)
            _valid = _validate_path(lyr, _fullpath)
            if _valid:
                _objpaths.append(_fullpath)
            else:
                print "invalid path"
        _routes[_obj] = _objpaths

def _make_paths(pathdict, maxlen, tail, path):
    _paths = []
    _pathlen = len(path)
    _head = None
    if _pathlen <= maxlen and tail in pathdict:
        if _pathlen:
            _head = path[0]
        for _next in pathdict[tail]:
            if _next is _head:
                _good = True
                if _pathlen == 2:
                    if (isinstance(path[0], segment.Segment) and
                        isinstance(path[1], segment.Segment)):
                        _good = False
                if _good:
                    _paths.append(path)
            elif _next not in path:
                _path = path + [_next]
                for _newpath in _make_paths(pathdict, maxlen, _next, _path):
                    _paths.append(_newpath)
            else:
                pass
    return _paths

hpx = 3.0
hpy = 4.0

def hatchtests():
    p1 = point.Point(0,0)
    p2 = point.Point(10,0)
    p3 = point.Point(10,10)
    p4 = point.Point(0,10)
    s1 = segment.Segment(p1,p2)
    s2 = segment.Segment(p2,p3)
    s3 = segment.Segment(p3,p4)
    s4 = segment.Segment(p4,p1)
    l1 = layer.Layer('foo')
    l1.addObject(p1)
    l1.addObject(p2)
    l1.addObject(p3)
    l1.addObject(p4)
    l1.addObject(s1)
    l1.addObject(s2)
    l1.addObject(s3)
    l1.addObject(s4)
    # find_hatched_area(l1, hpx, hpy)
    p5 = point.Point(2.95, 3.95)
    l1.addObject(p5)
    c1 = circle.Circle(p5, 1)
    l1.addObject(c1)
    # find_hatched_area(l1, hpx, hpy)
    
if __name__ == '__main__':
    hatchtests()



#
# Copyright (c) 2002, 2003, 2004 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# functions for doing modifications on drawing
# entities
#

import pygtk
pygtk.require('2.0')
import gtk

import Generic.dimension
import Generic.color
import Generic.graphicobject
import Generic.segment
import Generic.circle
import Generic.arc
import Generic.polyline

import Generic.move
import Generic.transfer
import Generic.split
import Generic.delete

#
# common code
#

def select_motion_notify(gtkimage, widget, event, tool):
    _tx, _ty = tool.getLocation()
    _px, _py = gtkimage.coordToPixTransform(_tx, _ty)
    _gc = gtkimage.getGC()
    _x = int(event.x)
    _y = int(event.y)
    _cp = tool.getCurrentPoint()
    if _cp is not None:
        _xc, _yc = _cp
        _xmin = min(_xc, _px)
        _ymin = min(_yc, _py)
        _rw = abs(_xc - _px)
        _rh = abs(_yc - _py)
        widget.window.draw_rectangle(_gc, gtk.FALSE, _xmin, _ymin, _rw, _rh)
    tool.setCurrentPoint(_x, _y)
    _xmin = min(_x, _px)
    _ymin = min(_y, _py)
    _rw = abs(_x - _px)
    _rh = abs(_y - _py)
    widget.window.draw_rectangle(_gc, gtk.FALSE, _xmin, _ymin, _rw, _rh)

def entry_get_distance(text):
    try:
        _text_dist = eval(text)
    except:
        raise ValueError, "Invalid text: " + str(text)
    _dist = _text_dist
    if not isinstance(_dist, float):
        _dist = float(_text_dist)
    return _dist
            
#
# move objects
#

def move_objects(gtkimage, objlist, tool):
    _init_func = tool.getHandler("initialize")
    _active_layer = gtkimage.getActiveLayer()
    _dx, _dy = tool.getDistance()
    gtkimage.startAction()
    try:
        Generic.move.move_objects(_active_layer, objlist, _dx, _dy)
    finally:
        gtkimage.endAction()
    gtkimage.redraw()
    gtkimage.setPrompt("Click in the drawing area or enter a distance.")
    tool.reset()
    _init_func(tool)
    
def move_button_press(gtkimage, tool):
    _x, _y = gtkimage.getPoint()
    #
    # need to find if the point is an intersection of drawing objects ...
    #
    tool.pushObject(_x)
    tool.pushObject(_y)

def move_end_button_press_cb(gtkimage, widget, event, tool):
    _x2, _y2 = gtkimage.getPoint()
    _x1, _y1 = tool.getLocation()
    _xmin = min(_x1, _x2)
    _xmax = max(_x1, _x2)
    _ymin = min(_y1, _y2)
    _ymax = max(_y1, _y2)
    _active_layer = gtkimage.getActiveLayer()
    _objlist = _active_layer.objsInRegion(_xmin, _ymin, _xmax, _ymax)
    move_objects(gtkimage, _objlist, tool)

def move_elem_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _objdict = gtkimage.mapPoint(_x, _y, _tol, None)
    if len(_objdict):
        _active_layer = gtkimage.getActiveLayer()
        if _active_layer in _objdict:
            _objs = []
            for _obj, _pt in _objdict[_active_layer]:
                _objs.append(_obj)
            _dx, _dy = tool.getDistance()
            move_objects(gtkimage, _objs, tool)
    else:
        _pt, _flag = gtkimage.findPoint(_x, _y, _tol)
        if _pt is not None:
            _x, _y = _pt.getCoords()
        tool.setLocation(_x, _y)
        tool.setHandler("motion_notify", select_motion_notify)
        tool.setHandler("button_press", move_end_button_press_cb)
        _gc = gtkimage.getGC()
        _gc.set_line_attributes(1, gtk.gdk.LINE_SOLID,
                                gtk.gdk.CAP_BUTT, gtk.gdk.JOIN_MITER)
        _gc.set_function(gtk.gdk.INVERT)
    
#
# move horizontal
#

def move_horizontal_entry_event(gtkimage, widget, tool):
    _entry = gtkimage.getEntry()
    _text = _entry.get_chars(0,-1)
    _entry.delete_text(0,-1)
    _dist = entry_get_distance(_text)
    tool.setDistance(_dist, 0.0)
    gtkimage.setPrompt("Click on the objects to move.")
    tool.setHandler("button_press", move_elem_button_press_cb)
    tool.delHandler("entry_event")

def move_horizontal_second_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _pt, _flag = gtkimage.findPoint(_x, _y, _tol)
    if _pt is not None:
        _x, _y = _pt.getCoords()
    _x1, _y1 = tool.getLocation()
    tool.setDistance((_x - _x1), 0.0)
    tool.clearLocation()
    gtkimage.setPrompt("Select the objects to move.")
    tool.setHandler("button_press", move_elem_button_press_cb)
    tool.delHandler("entry_event")

def move_horizontal_first_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _pt, _flag = gtkimage.findPoint(_x, _y, _tol)
    if _pt is not None:
        _x, _y = _pt.getCoords()
    tool.setLocation(_x, _y)
    gtkimage.setPrompt("Click another point to define the distance")
    tool.setHandler("button_press", move_horizontal_second_button_press_cb)
    
def move_horizontal_init(tool):
    tool.setHandler("button_press", move_horizontal_first_button_press_cb)
    tool.setHandler("entry_event", move_horizontal_entry_event)
    tool.setHandler("initialize", move_horizontal_init)

#
# move vertical
#

def move_vertical_entry_event(gtkimage, widget, tool):
    _entry = gtkimage.getEntry()
    _text = _entry.get_chars(0,-1)
    _entry.delete_text(0,-1)
    _dist = entry_get_distance(_text)
    tool.setDistance(0.0, _dist)
    gtkimage.setPrompt("Click on the objects to move.")
    tool.setHandler("button_press", move_elem_button_press_cb)
    tool.delHandler("entry_event")
    
def move_vertical_second_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _pt, _flag = gtkimage.findPoint(_x, _y, _tol)
    if _pt is not None:
        _x, _y = _pt.getCoords()
    _x1, _y1 = tool.getLocation()
    tool.setDistance(0.0, (_y - _y1))
    tool.clearLocation()
    gtkimage.setPrompt("Select the objects to move.")
    tool.setHandler("button_press", move_elem_button_press_cb)
    tool.delHandler("entry_event")

def move_vertical_first_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _pt, _flag = gtkimage.findPoint(_x, _y, _tol)
    if _pt is not None:
        _x, _y = _pt.getCoords()
    tool.setLocation(_x, _y)
    gtkimage.setPrompt("Click another point to define the distance")
    tool.setHandler("button_press", move_vertical_second_button_press_cb)

def move_vertical_init(tool):
    tool.setHandler("button_press", move_vertical_first_button_press_cb)
    tool.setHandler("entry_event", move_vertical_entry_event)
    tool.setHandler("initialize", move_vertical_init)

#
# move based on two mouse clicks
#
# it would be good to add an entry event handler here ...
#

def move_twopoint_second_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _pt, _flag = gtkimage.findPoint(_x, _y, _tol)
    if _pt is not None:
        _x, _y = _pt.getCoords()
    _x1, _y1 = tool.getLocation()
    tool.setDistance((_x - _x1), (_y - _y1))
    tool.clearLocation()
    gtkimage.setPrompt("Select the objects to move.")
    tool.setHandler("button_press", move_elem_button_press_cb)
    tool.delHandler("entry_event")

def move_twopoint_first_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _pt, _flag = gtkimage.findPoint(_x, _y, _tol)
    if _pt is not None:
        _x, _y = _pt.getCoords()
    tool.setLocation(_x, _y)
    gtkimage.setPrompt("Click another point to define the distance")
    tool.setHandler("button_press", move_twopoint_second_button_press_cb)

def move_twopoint_init(tool):
    tool.setHandler("button_press", move_twopoint_first_button_press_cb)
    tool.setHandler("initialize", move_twopoint_init)

#
# delete objects
#

def delete_region_end_cb(gtkimage, widget, event, tool):
    _x2, _y2 = gtkimage.getPoint()
    _x1, _y1 = tool.getLocation()
    _xmin = min(_x1, _x2)
    _xmax = max(_x1, _x2)
    _ymin = min(_y1, _y2)
    _ymax = max(_y1, _y2)
    tool.delHandler("motion_notify")
    tool.setHandler("button_press", delete_button_press_cb)
    _active_layer = gtkimage.getActiveLayer()
    _objs = _active_layer.objsInRegion(_xmin, _ymin, _xmax, _ymax)
    gtkimage.refresh()
    if len(_objs):
        gtkimage.startAction()
        try:
            Generic.delete.delete_objects(_active_layer, _objs)
        finally:
            gtkimage.endAction()
        tool.reset()
        gtkimage.redraw()
    delete_mode_init(tool)
    
def delete_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _active_layer = gtkimage.getActiveLayer()
    _tol = gtkimage.getTolerance()
    _objs = _active_layer.mapPoint((_x, _y), _tol)
    if len(_objs):
        _dims = []
        gtkimage.startAction()
        try:
            for _obj in _objs:
                if isinstance(_obj, Generic.dimension.Dimension):
                    _dims.append(_obj)
                elif isinstance(_obj, tuple):
                    _entity, _pt = _obj
                    _active_layer.delObject(_entity)
                else:
                    raise TypeError, "Unhandled object: " + `_obj`
            for _dim in _dims:
                if _dim in _active_layer: # it may have been removed ...
                    _active_layer.delObject(_dim)
        finally:
            gtkimage.endAction()
        gtkimage.redraw()
    else:
        tool.setLocation(_x, _y)
        tool.setHandler("motion_notify", select_motion_notify)
        tool.setHandler("button_press", delete_region_end_cb)
        _gc = gtkimage.getGC()
        _gc.set_line_attributes(1, gtk.gdk.LINE_SOLID,
                                gtk.gdk.CAP_BUTT, gtk.gdk.JOIN_MITER)
        _gc.set_function(gtk.gdk.INVERT)
        
def delete_mode_init(tool):
    tool.setHandler("button_press", delete_button_press_cb)
    tool.setHandler("initialize", delete_mode_init)

#
# stretch operations
#

def stretch_end_button_press_cb(gtkimage, widget, event, tool):
    _x2, _y2 = gtkimage.getPoint()
    _y1 = tool.popObject()
    _x1 = tool.popObject()
    _xmin = min(_x1, _x2)
    _xmax = max(_x1, _x2)
    _ymin = min(_y1, _y2)
    _ymax = max(_y1, _y2)
    tool.delHandler("motion_notify")
    _active_layer = gtkimage.getActiveLayer()
    _dx, _dy = tool.getDistance()
    for _pt in _active_layer.getLayerEntities("point"):
        if _pt.x > _xmax:
            break
        if _pt.inRegion(_xmin, _ymin, _xmax, _ymax):
            _pt.move(_dx, _dy)
    gtkimage.redraw()
    tool.clearLocation()
    tool.clearCurrentPoint()
    tool.setHandler("button_press", stretch_elem_button_press_cb)

def stretch_elem_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _active_layer = gtkimage.getActiveLayer()
    _pt = _active_layer.find('point', _x, _y, _tol)
    if _pt is not None:
        _dx, _dy = tool.getDistance()
        _pt.move(_dx, _dy)
        gtkimage.redraw()
    else:
        tool.pushObject(_x)
        tool.pushObject(_y)
        tool.setLocation(_x, _y)
        tool.setHandler("motion_notify", select_motion_notify)
        tool.setHandler("button_press", stretch_end_button_press_cb)
        _gc = gtkimage.getGC()
        _gc.set_line_attributes(1, gtk.gdk.LINE_SOLID,
                                gtk.gdk.CAP_BUTT, gtk.gdk.JOIN_MITER)
        _gc.set_function(gtk.gdk.INVERT)

#
# stretch horizontal
#

def stretch_horiz_button_press_cb(gtkimage, widget, event, tool):
    if not len(tool):
        move_button_press(gtkimage, tool)
        gtkimage.setPrompt("Click a second point to indicate the horizontal distance")
    else:
        _x, _y = gtkimage.getPoint()
        #
        # see if the point is at an intersection of drawing objects ...
        #
        _y1 = tool.popObject()
        _x1 = tool.popObject()
        tool.setDistance((_x - _x1), 0.0)
        gtkimage.setPrompt("Select the points to move.")
        # tool.delHandler("entry_event")
        tool.setHandler("button_press", stretch_elem_button_press_cb)

def stretch_horizontal_init(tool):
    tool.initialize()
    tool.setHandler("button_press", stretch_horiz_button_press_cb)
    # tool.setHandler("entry_event", stretch_horiz_entry_event)

#
# stretch vertical
#

def stretch_vert_button_press_cb(gtkimage, widget, event, tool):
    if not len(tool):
        move_button_press(gtkimage, tool)
        gtkimage.setPrompt("Click a second point to indicate the vertical distance")
    else:
        _x, _y = gtkimage.getPoint()
        #
        # see if the point is at an intersection of drawing objects ...
        #
        _y1 = tool.popObject()
        _x1 = tool.popObject()
        tool.setDistance(0.0, (_y - _y1))
        gtkimage.setPrompt("Select the points to move.")
        # tool.delHandler("entry_event")
        tool.setHandler("button_press", stretch_elem_button_press_cb)

def stretch_vertical_init(tool):
    tool.initialize()
    tool.setHandler("button_press", stretch_vert_button_press_cb)
    # tool.setHandler("entry_event", stretch_horiz_entry_event)

#
# split objects into two pieces or at intersection points
#

def split_end_button_press_cb(gtkimage, widget, event, tool):
    _x2, _y2 = gtkimage.getPoint()
    _y1 = tool.popObject()
    _x1 = tool.popObject()
    _xmin = min(_x1, _x2)
    _xmax = max(_x1, _x2)
    _ymin = min(_y1, _y2)
    _ymax = max(_y1, _y2)
    tool.delHandler("motion_notify")
    _active_layer = gtkimage.getActiveLayer()
    _objs = _active_layer.objsInRegion(_xmin, _ymin, _xmax, _ymax, True)
    gtkimage.refresh()
    if len(_objs):
        _splitable = []
        for _obj in _objs:
            if isinstance(_obj, (Generic.segment.Segment,
                                 Generic.circle.Circle,
                                 Generic.arc.Arc,
                                 Generic.polyline.Polyline)):
                _splitable.append(_obj)
        if len(_splitable):
            gtkimage.startAction()
            try:
                Generic.split.split_objects(_active_layer, _splitable)
            finally:
                gtkimage.endAction()
        gtkimage.redraw()
    gtkimage.setPrompt("Click on the objects you want to split.")
    tool.clearLocation()
    tool.clearCurrentPoint()
    tool.setHandler("button_press", split_object_button_press_cb)
    
def split_object_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _active_layer = gtkimage.getActiveLayer()
    _objlist = _active_layer.mapPoint((_x, _y), _tol, None)
    if len(_objlist):
        _redraw = False
        for _obj, _pt in _objlist:
            if isinstance(_obj, Generic.segment.Segment):
                _redraw = True
                _p1, _p2 = _obj.getEndpoints()
                _lpt = _active_layer.find('point', _pt.x, _pt.y)
                if _lpt is None:
                    _active_layer.addObject(_pt)
                    _lpt = _pt
                _s1, _s2 = Generic.split.split_segment(_obj, _pt)
                gtkimage.startAction()
                try:
                    _active_layer.addObject(_s1)
                    _active_layer.addObject(_s2)
                    _active_layer.delObject(_obj)
                finally:
                    gtkimage.endAction()
            elif isinstance(_obj, Generic.arc.Arc):
                _redraw = True
                _arc1, _arc2 = Generic.split.split_arc(_obj, _pt)
                gtkimage.startAction()
                try:
                    _active_layer.addObject(_arc1)
                    _active_layer.addObject(_arc2)
                    _active_layer.delObject(_obj)
                finally:
                    gtkimage.endAction()
            elif isinstance(_obj, Generic.circle.Circle):
                _redraw = True
                _arc = Generic.split.split_circle(_obj, _pt)
                gtkimage.startAction()
                try:
                    _active_layer.addObject(_arc)
                    _active_layer.delObject(_obj)
                finally:
                    gtkimage.endAction()
            elif isinstance(_obj, Generic.polyline.Polyline):
                _redraw = True
                _lpt = _active_layer.find('point', _pt.x, _pt.y)
                gtkimage.startAction()
                try:
                    if _lpt is None:
                        _active_layer.addObject(_pt)
                        _lpt = _pt
                    Generic.split.split_polyline(_obj, _lpt)
                finally:
                    gtkimage.endAction()
            else:
                pass
        if _redraw:
            gtkimage.redraw()
    else:
        tool.pushObject(_x)
        tool.pushObject(_y)
        tool.setLocation(_x, _y)
        tool.setHandler("motion_notify", select_motion_notify)
        tool.setHandler("button_press", split_end_button_press_cb)
        _gc = gtkimage.getGC()
        _gc.set_line_attributes(1, gtk.gdk.LINE_SOLID,
                                gtk.gdk.CAP_BUTT, gtk.gdk.JOIN_MITER)
        _gc.set_function(gtk.gdk.INVERT)
        
def split_object_init(tool):
    tool.initialize()
    tool.setHandler("button_press", split_object_button_press_cb)
    tool.setHandler("initialize", split_object_init)

#
# transfer objects from one layer to another
#

def transfer_end_button_press_cb(gtkimage, widget, event, tool):
    _x2, _y2 = gtkimage.getPoint()
    _y1 = tool.popObject()
    _x1 = tool.popObject()
    _xmin = min(_x1, _x2)
    _xmax = max(_x1, _x2)
    _ymin = min(_y1, _y2)
    _ymax = max(_y1, _y2)
    tool.delHandler("motion_notify")
    _active_layer = gtkimage.getActiveLayer()
    _layers = [gtkimage.getTopLayer()]
    _objdict = {}
    while len(_layers):
        _layer = _layers.pop()
        if _layer is not _active_layer:
            if _layer.isVisible():
                _objs = _layer.objsInRegion(_xmin, _ymin, _xmax, _ymax)
                if len(_objs):
                    _objdict[_layer] = _objs
        _layers.extend(_layer.getSublayers())
    if len(_objdict):
        gtkimage.startAction()
        try:
            for _layer in _objdict:
                if _layer is not _active_layer:
                    _objs = _objdict[_layer]
                    Generic.transfer.transfer_objects(_objs, _layer, _active_layer)
        finally:
            gtkimage.endAction()
        gtkimage.redraw()
    tool.clearLocation()
    tool.clearCurrentPoint()
    tool.setHandler("button_press", transfer_object_button_press_cb)
    
def transfer_object_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _active_layer = gtkimage.getActiveLayer()
    _objdict = gtkimage.mapPoint(_x, _y, _tol, None)
    if len(_objdict):
        gtkimage.startAction()
        try:
            for _layer in _objdict:
                if _layer is not _active_layer:
                    _objs = []
                    for _obj, _pt in _objdict[_layer]:
                        _objs.append(_obj)
                    Generic.transfer.transfer_objects(_objs, _layer, _active_layer)
        finally:
            gtkimage.endAction()
        gtkimage.redraw()
    else:
        tool.pushObject(_x)
        tool.pushObject(_y)
        tool.setLocation(_x, _y)
        tool.setHandler("motion_notify", select_motion_notify)
        tool.setHandler("button_press", transfer_end_button_press_cb)
        gc = gtkimage.getGC()
        gc.set_line_attributes(1, gtk.gdk.LINE_SOLID,
                               gtk.gdk.CAP_BUTT, gtk.gdk.JOIN_MITER)
        gc.set_function(gtk.gdk.INVERT)
        
def transfer_object_init(tool):
    tool.initialize()
    tool.setHandler("button_press", transfer_object_button_press_cb)
    
#
# change color
#

def change_color_init(gtkimage):
    _window = gtkimage.getWindow()
    _dialog = gtk.ColorSelectionDialog("Set Active Color")
    _dialog.set_transient_for(_window)
    _colorsel = _dialog.colorsel

    _prev_color = gtkimage.getOption('LINE_COLOR')
    if gtkimage.hasGTKColor(_prev_color):
        _gtk_color = gtkimage.getGTKColor(_prev_color)
    else:
        _da = gtkimage.getDA()
        _cstring = str(_prev_color)
        _gtk_color = _da.get_colormap().alloc_color(_cstring)
        gtkimage.saveGTKColor(_prev_color, _gtk_color)
    _colorsel.set_previous_color(_gtk_color)
    _colorsel.set_current_color(_gtk_color)
    _colorsel.set_has_palette(gtk.TRUE)
    _response = _dialog.run()
    if _response == gtk.RESPONSE_OK:
        _gtk_color = _colorsel.get_current_color()
        _red = int((_gtk_color.red/65535.0) * 255.0)
        _green = int((_gtk_color.green/65535.0) * 255.0)
        _blue = int((_gtk_color.blue/65535.0) * 255.0)
        _col = Generic.color.get_color(_red, _green, _blue)
        if gtkimage.hasSelection():
            _active_layer = gtkimage.getActiveLayer()
            _objs = []
            for _layer, _obj in gtkimage.getSelectedObjects():
                if _layer is _active_layer:
                    if isinstance(_obj, Generic.graphicobject.GraphicObject):
                        _objs.append(_obj)
            if len(_objs):
                gtkimage.startAction()
                try:
                    for _obj in _objs:
                        _obj.setColor(_col)
                finally:
                    gtkimage.endAction()
    _dialog.destroy()
    gtkimage.redraw()

#
# change linetypes
#

def change_linetype_init(gtkimage):
    _window = gtkimage.getWindow()
    _dialog = gtk.Dialog("Change Linetype", _window,
                         gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                         (gtk.STOCK_OK, gtk.RESPONSE_OK,
                          gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))
    _hbox = gtk.HBox(gtk.FALSE, 2)
    _hbox.set_border_width(2)
    _dialog.vbox.pack_start(_hbox, gtk.FALSE, gtk.FALSE, 0)
    
    _label = gtk.Label("Linetype:")
    _hbox.pack_start(_label, gtk.FALSE, gtk.FALSE, 0)
    _clt = gtkimage.getOption('LINE_TYPE')
    _linetypes = gtkimage.getAllLinetypes()
    _menu = gtk.Menu()
    _idx = 0
    for _i in range(len(_linetypes)):
        _lt = _linetypes[_i]
        if _lt is _clt:
            _idx = _i
        _item = gtk.MenuItem(_lt.getName())
        _menu.append(_item)
    _opt_menu = gtk.OptionMenu()
    _opt_menu.set_menu(_menu)
    _opt_menu.set_history(_idx)
    _hbox.pack_start(_opt_menu, gtk.TRUE, gtk.TRUE, 0)
    _dialog.show_all()
    _response = _dialog.run()
    if _response == gtk.RESPONSE_OK:
        _lt = _linetypes[_opt_menu.get_history()]
        if gtkimage.hasSelection():
            _active_layer = gtkimage.getActiveLayer()
            _objs = []
            for _layer, _obj in gtkimage.getSelectedObjects():
                if _layer is _active_layer:
                    if isinstance(_obj, Generic.graphicobject.GraphicObject):
                        _objs.append(_obj)
            if len(_objs):
                gtkimage.startAction()
                try:
                    for _obj in _objs:
                        _obj.setLinetype(_lt)
                finally:
                    gtkimage.endAction()
    _dialog.destroy()
    gtkimage.redraw()

#
# change thickness
#

def change_thickness_init(gtkimage):
    _window = gtkimage.getWindow()
    _dialog = gtk.Dialog("Change Thickness", _window,
                         gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                         (gtk.STOCK_OK, gtk.RESPONSE_OK,
                          gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))
    _hbox = gtk.HBox(gtk.FALSE, 2)
    _hbox.set_border_width(2)
    _dialog.vbox.pack_start(_hbox, gtk.FALSE, gtk.FALSE, 0)
    _label = gtk.Label("Thickness:")
    _hbox.pack_start(_label, gtk.FALSE, gtk.FALSE, 0)
    _thickness = gtkimage.getOption('LINE_THICKNESS')
    _adj = gtk.Adjustment(_thickness, 0.0001, 20.0, 0.1, 1.0, 1.0)
    _sb = gtk.SpinButton(_adj)
    _sb.set_digits(1)
    _sb.set_numeric(gtk.FALSE)
    _hbox.pack_start(_sb, gtk.TRUE, gtk.TRUE, 0)
    _dialog.show_all()
    _response = _dialog.run()
    if _response == gtk.RESPONSE_OK:
        _nt = float(_sb.get_value())
        if gtkimage.hasSelection():
            _active_layer = gtkimage.getActiveLayer()
            _objs = []
            for _layer, _obj in gtkimage.getSelectedObjects():
                if _layer is _active_layer:
                    if isinstance(_obj, Generic.graphicobject.GraphicObject):
                        _objs.append(_obj)
            if len(_objs):
                gtkimage.startAction()
                try:
                    for _obj in _objs:
                        _obj.setThickness(_nt)
                finally:
                    gtkimage.endAction()
    _dialog.destroy()
    gtkimage.redraw()

#
# change the style
#

def change_style_init(gtkimage):
    _window = gtkimage.getWindow()
    _dialog = gtk.Dialog("Change Style", _window,
                         gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                         (gtk.STOCK_OK, gtk.RESPONSE_OK,
                          gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))
    _hbox = gtk.HBox(gtk.FALSE, 2)
    _hbox.set_border_width(2)
    _dialog.vbox.pack_start(_hbox, gtk.FALSE, gtk.FALSE, 0)
    _label = gtk.Label("Style:")
    _cst = gtkimage.getOption('LINE_STYLE')
    _styles = gtkimage.getAllStyles()
    _menu = gtk.Menu()
    _idx = 0
    for _i in range(len(_styles)):
        _s = _styles[_i]
        if _s is _cst:
            _idx = _i
        _item = gtk.MenuItem(_s.getName())
        _menu.append(_item)
    _opt_menu = gtk.OptionMenu()
    _opt_menu.set_menu(_menu)
    _opt_menu.set_history(_idx)
    _hbox.pack_start(_label, gtk.FALSE, gtk.FALSE, 0)
    _hbox.pack_start(_opt_menu, gtk.TRUE, gtk.TRUE, 0)
    _dialog.show_all()
    _response = _dialog.run()
    if _response == gtk.RESPONSE_OK:
        _s = _styles[_opt_menu.get_history()]
        if gtkimage.hasSelection():
            _active_layer = gtkimage.getActiveLayer()
            _objs = []
            for _layer, _obj in gtkimage.getSelectedObjects():
                if _layer is _active_layer:
                    if isinstance(_obj, Generic.graphicobject.GraphicObject):
                        _objs.append(_obj)
            if len(_objs):
                gtkimage.startAction()
                try:
                    for _obj in _objs:
                        _obj.setStyle(_s)
                finally:
                    gtkimage.endAction()
    _dialog.destroy()
    gtkimage.redraw()

#
# arbitrary zoom
#

def zoom_end_button_press_cb(gtkimage, widget, event, tool):
    _xp, _yp = gtkimage.getPoint()
    _x1, _y1 = tool.getLocation()
    _xmin = min(_xp, _x1)
    _ymin = min(_yp, _y1)
    _width, _height = gtkimage.getSize()
    _fw = float(_width)
    _fh = float(_height)
    _wpp = abs(_x1 - _xp)/_fw
    _hpp = abs(_y1 - _yp)/_fh
    if _wpp > _hpp:
        _scale = _wpp
    else:
        _scale = _hpp
    gtkimage.setView(_xmin, _ymin, _scale)
    gtkimage.setPrompt("Click in the drawing.")
    gtkimage.redraw()
    zoom_init(tool)

def zoom_motion_notify(gtkimage, widget, event, tool):
    _tx, _ty = tool.getLocation()
    _px, _py = gtkimage.coordToPixTransform(_tx, _ty)
    # width, height = gtkimage.getSize()
    _gc = gtkimage.getGC()
    _x = int(event.x)
    _y = int(event.y)
    _cp = tool.getCurrentPoint()
    #
    # it would be nice to draw the rectangle in the current
    # shape of the window ...
    #
    if _cp is not None:
        _xc, _yc = _cp
        _xmin = min(_px, _xc)
        _ymin = min(_py, _yc)
        _rw = abs(_xc - _px)
        _rh = abs(_yc - _py)
        widget.window.draw_rectangle(_gc, gtk.FALSE, _xmin, _ymin, _rw, _rh)
    _xmin = min(_x, _px)
    _ymin = min(_y, _py)
    tool.setCurrentPoint(_x, _y)
    _rw = abs(_x - _px)
    _rh = abs(_y - _py)
    widget.window.draw_rectangle(_gc, gtk.FALSE, _xmin, _ymin, _rw, _rh)

def zoom_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    tool.setLocation(_x, _y)
    tool.setHandler("motion_notify", zoom_motion_notify)
    tool.setHandler("button_press", zoom_end_button_press_cb)
    gtkimage.setPrompt("Click a second point to define the zoom window.")
    _gc = gtkimage.getGC()
    _gc.set_line_attributes(1, gtk.gdk.LINE_SOLID,
                            gtk.gdk.CAP_BUTT, gtk.gdk.JOIN_MITER)
    _gc.set_function(gtk.gdk.INVERT)

def zoom_init(tool):
    tool.initialize()
    tool.setHandler("button_press", zoom_button_press_cb)
